Package: coconots
Type: Package
Title: Convolution-Closed Models for Count Time Series
Version: 2.0.3
Date: 2026-06-14
Authors@R: c(
    person(given = "Manuel",
           family = "Huth",
           role = c("aut", "cre"),
           email = "manuel.huth@yahoo.com"),
    person(given = "Robert C.",
           family = "Jung",
           role = "aut"),
    person(given = "Andy",
           family = "Tremayne",
           role =  "aut"
           ))
Description: Useful tools for fitting, validating, and forecasting of practical convolution-closed time series models for low counts are provided. Marginal distributions of the data can be modelled via Poisson and Generalized Poisson innovations. Regression effects can be incorporated through time varying innovation rates. The models are described in Jung and Tremayne (2011) <doi:10.1111/j.1467-9892.2010.00697.x> and the model assessment tools are presented in Czado et al. (2009) <doi:10.1111/j.1541-0420.2009.01191.x> and, Tsay (1992) <doi:10.2307/2347612>.
LazyData: true
LinkingTo: Rcpp
RoxygenNote: 7.3.3
Depends: R (>= 4.0.2)
Imports: forecast, numDeriv, HMMpa, ggplot2, matrixStats,
        JuliaConnectoR, Rcpp, stats
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
License: MIT + file LICENSE
NeedsCompilation: yes
Packaged: 2026-06-14 10:18:02 UTC; manuel
Author: Manuel Huth [aut, cre],
  Robert C. Jung [aut],
  Andy Tremayne [aut]
Maintainer: Manuel Huth <manuel.huth@yahoo.com>
Repository: CRAN
Date/Publication: 2026-06-14 10:50:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-06-14 11:04:55 UTC; unix
Archs: coconots.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(autoplot,coco)
S3method(autoplot,cocoBoot)
S3method(autoplot,cocoForecast)
S3method(autoplot,cocoForecastCollection)
S3method(autoplot,cocoPit)
S3method(autoplot,cocoResid)
S3method(coef,coco)
S3method(extractAIC,coco)
S3method(fitted,coco)
S3method(fitted,cocoResid)
S3method(logLik,coco)
S3method(nobs,coco)
S3method(plot,coco)
S3method(plot,cocoBoot)
S3method(plot,cocoForecast)
S3method(plot,cocoForecastCollection)
S3method(plot,cocoPit)
S3method(plot,cocoResid)
S3method(predict,coco)
S3method(print,coco)
S3method(print,cocoBoot)
S3method(print,cocoForecastCollection)
S3method(print,cocoPit)
S3method(print,cocoResid)
S3method(print,summary.coco)
S3method(print,summary.cocoBoot)
S3method(print,summary.cocoForecast)
S3method(print,summary.cocoForecastCollection)
S3method(print,summary.cocoPit)
S3method(print,summary.cocoResid)
S3method(print,summary.cocoSoc)
S3method(residuals,coco)
S3method(residuals,cocoResid)
S3method(rstandard,coco)
S3method(summary,coco)
S3method(summary,cocoBoot)
S3method(summary,cocoForecast)
S3method(summary,cocoForecastCollection)
S3method(summary,cocoPit)
S3method(summary,cocoResid)
S3method(summary,cocoSoc)
S3method(vcov,coco)
export(autoplot)
export(cocoBoot)
export(cocoPit)
export(cocoReg)
export(cocoResid)
export(cocoScore)
export(cocoSim)
export(cocoSoc)
export(installJuliaPackages)
export(setJuliaSeed)
importFrom(HMMpa,dgenpois)
importFrom(JuliaConnectoR,juliaCall)
importFrom(JuliaConnectoR,juliaEval)
importFrom(JuliaConnectoR,juliaGet)
importFrom(JuliaConnectoR,juliaLet)
importFrom(Rcpp,sourceCpp)
importFrom(forecast,Acf)
importFrom(ggplot2,autoplot)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,theme_bw)
importFrom(matrixStats,rowQuantiles)
importFrom(numDeriv,grad)
importFrom(numDeriv,hessian)
importFrom(stats,coef)
importFrom(stats,constrOptim)
importFrom(stats,extractAIC)
importFrom(stats,fitted)
importFrom(stats,glm)
importFrom(stats,lm)
importFrom(stats,logLik)
importFrom(stats,na.omit)
importFrom(stats,nobs)
importFrom(stats,optim)
importFrom(stats,pchisq)
importFrom(stats,predict)
importFrom(stats,qchisq)
importFrom(stats,residuals)
importFrom(stats,rpois)
importFrom(stats,rstandard)
importFrom(stats,runif)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(utils,head)
useDynLib(coconots, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# coconots

<!-- badges: start -->
[![R-CMD-check](https://github.com/manuhuth/coconots/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/manuhuth/coconots/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/coconots)](https://CRAN.R-project.org/package=coconots)
[![Codecov test
coverage](https://codecov.io/gh/manuhuth/coconots/branch/main/graph/badge.svg)](https://app.codecov.io/gh/manuhuth/coconots?branch=main)
<!-- badges: end -->

Functions to analyse time series consisting of low counts are provided.
The focus in the current version is on practical models that can capture
first and higher-order dependence based on the work of Joe (1996). Both
equidispersed and overdispersed marginal distributions of data can be
modelled. Regression effects can be included. Fast and efficient
procedures for likelihood based inference and probabilistic forecasting
are provided as well as useful tools for model validation and
diagnostics.

<figure>
<img
src="https://github.com/manuhuth/coconots/blob/main/man/figures/functionality.png?raw=true"
alt="alt text" />
<figcaption aria-hidden="true">alt text</figcaption>
</figure>

## Details

The package allows simulation of convolution-closed count time series
models with the cocoSim function. Model fitting is performed with the
cocoReg routine. By passing a cocoReg-type object, cocoForecast computes
the one-step ahead forecasting distribution. cocoBoot, cocoPit,
cocoScore, and cocoResid provide routines for model assessment. The main
usage of the package is illustrated within the cocoReg function chapter.
For more details and examples of the functions see the respective
sections within this vignette.

By default, our functions make use of an RCPP implementation. However,
users with a running Julia installation can choose to call Julia in the
background to run their functions by specifying it in the R function
input. This option is particularly useful for the regression (), where a
complex likelihood function must be numerically evaluated to obtain
parameter estimates. By leveraging Julia’s automatic differentiation
capabilities, our functions can take advantage of numerical gradients,
leading to increased numerical stability and faster convergence.

Despite these advantages, we found that both the Julia and RCPP
implementations produced qualitatively similar results in all our tests.
As a result, we have decided to use the RCPP implementation as the
default option to make our package accessible to non-Julia users.

## Model

<figure>
<img
src="https://github.com/manuhuth/coconots/blob/main/man/figures/dgp.png?raw=true"
alt="alt text" />
<figcaption aria-hidden="true">alt text</figcaption>
</figure>

## Installation

You can install the latest stable version of coconots from
[CRAN](https://cran.r-project.org/) with:

``` r
install.packages("coconots")
```

or the latest version from GitHub with:

``` r
devtools::install_github("manuhuth/coconots")
```

## Example using RCPP implementation

Coconots runs on an RCPP and a Julia implementation. RCPP is the
default. The regression results can be summarized using the summary
function.

``` r
library(coconots)
length <- 500

pars <- c(1, 0.4, 0.2)
set.seed(12345)
data <- cocoSim(order = 1, type = "GP", par = pars,
                length = length)
coco <- cocoReg(order = 1, type = "GP", data = data)
#> Registered S3 method overwritten by 'quantmod':
#>   method            from
#>   as.zoo.data.frame zoo
summary(coco)
#> Coefficients:
#>          Estimate   Std. Error         t
#> lambda     1.0548       0.0930   11.3367
#> alpha      0.4201       0.0373   11.2508
#> eta        0.1609       0.0329    4.8958
#> 
#> Type: GP 
#> Order: 1 
#> 
#> Log-likelihood: -885.012
```

## Example using Julia implementation

In order to use the Julia functions, you need a running version of Julia
installed. You need to install the relevant Julia packages once.

``` r
coconots::installJuliaPackages()
#> Starting Julia ...
```

If installed, the Julia implementation can be addressed by setting the
julia argument to true. In this case we need to set the seed within the
simulation function in order to pass it to Julia.

``` r
library(coconots)
length <- 500

pars <- c(1, 0.3, 0.01, 0.2)
data <- cocoSim(order = 2, type = "Poisson", par = pars, length = length,
                julia = TRUE, julia_seed = 123)
coco <- cocoReg(order = 2, type = "Poisson", data = data, julia = TRUE)
```

## Model assessment

We provide different tools for model assessment. One can even use the
RCPP implementation for the regression and do model assessment with the
Julia implementation. If this is desired, one needs to specify in the
output of the regression that it should be Julia compatible. Note that
this is not necessary if the Julia option in the regression is true.

``` r
library(coconots)
length <- 500

pars <- c(1, 0.4)
set.seed(12345)
data <- cocoSim(order = 1, type = "Poisson", par = pars, length = length)
coco <- cocoReg(order = 1, type = "Poisson", data = data,
                julia_installed=TRUE)
(scores <- cocoScore(coco, julia = TRUE))
#> $log.score
#> [1] 1.501935
#> 
#> $quad.score
#> [1] -0.2532058
#> 
#> $rps.score
#> [1] 0.619761
pit <- cocoPit(coco, julia = TRUE)
plot(pit)
```

<img src="man/figures/README-example_assessment-1.png" width="100%" />

``` r
boot <- cocoBoot(coco, julia = TRUE)
plot(boot)
```

<img src="man/figures/README-example_assessment-2.png" width="100%" />

``` r
forecast <- predict(coco, k=3, julia = TRUE)
plot(forecast) #plot all k forecasts
```

<img src="man/figures/README-example_assessment-3.png" width="100%" />

``` r
plot(forecast[[1]]) #plot density of one-step ahead forecast
```

<img src="man/figures/README-example_assessment-4.png" width="100%" />

``` r
resid <- cocoResid(coco)
plot(resid)
```

<img src="man/figures/README-example_assessment-5.png" width="100%" />

## Compare Models

Model selection is facilitated by providing quick and easy access to the
scoring metrics.

``` r
library(coconots)
length <- 500

pars <- c(1.3, 0.25, 0.03, 0.2, 0.3)
set.seed(12345)
data <- cocoSim(order = 2, type = "GP", par = pars, length = length)
soc <- cocoSoc(data, julia=T)
#> [1] "Computing PAR1 model."
#> [1] "Computing PAR2 model."
#> [1] "Computing GP1 model."
#> [1] "Computing GP2 model."
summary(soc)
#>            Logarithmic Score   Quadratic Score   Ranked Probability Score
#> 1   PAR1            2.265970        -0.1221436                   1.302572
#> 2   PAR2            2.264753        -0.1223451                   1.301770
#> 3    GP1            2.164318        -0.1363304                   1.263041
#> 4    GP2            2.161843        -0.1364925                   1.261362
```

pGP1 <- function(q, y, par) { # q is number we want to know the quantile of
  if ((round(y) != y) | (y < 0)) {
    stop("y must be a non-negative integer")
  }

  if (length(par) != 3) {
    stop("three parametes, lambda, alpha and eta, must be passed to the function")
  }

  q <- floor(q)

  lambda <- par[1]
  alpha <- par[2]
  eta <- par[3]

  if (lambda < 0) {
    stop("lambda must be greater than zero")
  }

  if ((alpha < 0) | (alpha > 1) | (eta < 0) | (eta > 1)) {
    stop("alpha and eta must be between or equal to zero and one")
  }

  CDF <- 0
  if (q >= 0) {
    for (j in 0:q) {
      CDF <- CDF + dGP1(j, y, par)
    }
  }


  return(CDF)
}

#' @importFrom stats runif
#' @importFrom HMMpa dgenpois
rgenpois <- function(n, lambda, eta){
  vec <- numeric(n)

  for (j in 1:n) {
    index <- 0
    unif <- stats::runif(1)
    nain <- HMMpa::dgenpois(0, lambda1 = lambda, lambda2 = eta)
    while(nain < unif) {
      index <- index + 1
      nain <- nain + HMMpa::dgenpois(index, lambda1 = lambda, lambda2 = eta)
    }
    vec[j] <- index
  }
  return(vec)
}

transformJuliaRegOutputToR <- function(xreg, pars, grad, hes, inv_hes, se,
                                       data, type, order, likelihood, end_time, start_time, julia_reg){
  if (is.null(xreg)){
    par_use <- c(pars[length(pars)], pars[1:(length(pars)-1)])
    se_use <- c(se[length(se)], se[1:(length(se)-1)])
  } else {
    par_use <- pars
    se_use <- se
  }
  
  return(list(
    "par" = par_use, "gradient" = grad, cov=xreg,
    "hessian" = hes, "inv hessian" = inv_hes, "se" = se_use, "ts" = data,
    "type" = type, "order" = order, "seasonality" = c(1,2), "likelihood" = likelihood,
    "duration" = end_time - start_time, julia_reg = julia_reg
  ))
}

#' Time Series of Monthly Counts of Claimants Collecting Wage Loss Benefit
#'
#' Monthly counts of claimants collecting wage loss benefit for injuries in the workplace 
#' at one specific service delivery location of the Workers Compensation Board 
#' of British Columbia, Canada in the period January 1985 to December 1994. 
#' Only injuries due to cuts and lacerations are considered. 
#' The data have been provided by Brendan McCabe.
#'
#' @source Freeland, R. K.; McCabe, B.P.M. (2004) 
#'   Analysis of count data by means of the Poisson autoregressive model. 
#'   \emph{Journal of Time Series Analysis}, \bold{25}, 701--722.
#'
#' @format A time series (`ts`) object containing monthly data from January 1985 to December 1994.
#' @keywords datasets
#' 
"cuts"

#' @title Fitting First- and Second Order (G)PAR Models 
#' @description The function fits first- and second-order (Generalized) Poisson
#' integer autoregressive \code{(G)PAR} time series models for count data as discussed in Jung and Tremayne (2011). Autoregressive dependence on past counts is modeled using a special random operator that preserves integer values and, through closure under convolution, ensures that the marginal distribution remains within the same family as the innovations.
#'
#' These models can be viewed as stationary finite-order Markov chains, where the innovation distribution can be either Poisson or Generalized Poisson, the latter accounting for overdispersion. Estimation is performed via maximum likelihood, with an option to impose linear constraints. Without constraints, parameters may fall outside the theoretically feasible space, but optimization may be faster.
#'
#' Method of moments estimators are used to initialize numerical optimization, though custom starting values can be provided. If \proglang{julia} is installed, users can opt to run the optimization in \proglang{julia} for potentially faster computation and improved numerical stability via automatic differentiation. See below for details on the \proglang{julia} implementation.
#' @param type character, either "Poisson" or "GP" indicating the type of the innovation distribution
#' @param order integer, either 1 or 2 indicating the order of the model
#' @param data time series data to be used in the analysis
#' @param xreg optional matrix of explanatory variables (without constant term) for use in a regression model
#' @param constrained.optim logical indicating whether optimization should be constrained, currently only available in the R version
#' @param b.beta numeric value indicating the lower bound for the parameters of the explanatory variables for the optimization, currently only available in the \proglang{R} version
#' @param start optional numeric vector of starting values for the optimization
#' @param start.val.adjust logical indicating whether starting values should be adjusted, currently only available in the R version
#' @param method_optim character string indicating the optimization method to be used, currently only available in the R version. In the \proglang{julia} implementation this is by default the LBFGS algorithm
#' @param replace.start.val numeric value indicating the value to replace any invalid starting values, currently only available in the \proglang{R} version
#' @param iteration.start.val numeric value indicating the proportion of the interval to use as the new starting value, currently only available in the \proglang{R} version
#' @param method.hessian character string indicating the method to be used to approximate the Hessian matrix, currently only available in the \proglang{R} version
#' @param cores numeric indicating the number of cores to use, currently only available in the \proglang{R} version (default: 2)
#' @param julia if TRUE, the model is estimated with \proglang{julia}. This can improve computational speed significantly since \proglang{julia} makes use of derivatives using autodiff. In this case, only \code{type}, \code{order}, \code{data}, \code{xreg}, and \code{start} are used as other inputs (default: FALSE).
#' @param julia_installed if TRUE, the model \proglang{R} output will contain a \proglang{julia} compatible output element.
#' @param link_function Specifies the link function for the conditional mean of the innovation (\eqn{\lambda}). The default is `log`, but other available options include `identity`, `relu`, and `softplus`. This parameter is applicable only when covariates are used. Note that using the `identity` link function may result in \eqn{\lambda} becoming negative. To prevent this, ensure all covariates are positive and restrict the parameter \eqn{\beta} to positive values by setting `b.beta` to a small positive value.
#' @importFrom JuliaConnectoR juliaGet juliaLet
#' @importFrom stats nobs rstandard vcov coef residuals fitted extractAIC logLik
#' @importFrom ggplot2 autoplot
#' @author Manuel Huth
#' @return an object of class coco. It contains the parameter estimates, standard errors, the log-likelihood, 
#' and information on the model specifications. If \proglang{julia} is used for parameter estimation or the \proglang{julia} installation
#' parameter is set to TRUE, the results contain an additional Julia element that is called from the model \proglang{julia}
#' assessment tools if they are run with the \proglang{julia} implementation.
#' @details 
#' Let a time series of counts be \eqn{\{X_t\}} and be \eqn{R(\cdot)} a random operator that differs between model specifications.
#' For more details on the random operator, see Jung and Tremayne (2011) and Joe (1996).
#' The general first-order model is of the form
#' \deqn{X_t = R(X_{t-1}) + W_t,}
#' and the general second-order model of the form
#' \deqn{X_t = R(X_{t-1}, X_{t-2}) + W_t,}
#' where \eqn{W_t} are i.i.d Poisson (\eqn{W_t \sim Po(\lambda_t)}) or Generalized
#' Poisson (\eqn{W_t \sim GP(\lambda_t, \eta)}) innovations. Through closure under convolution
#' the marginal distributions of \eqn{\{X_t\}} are therefore Poisson or Generalized Poisson distributions, respectively.
#' 
#' If no covariates are used \eqn{\lambda_t = \lambda} and if covariates are used
#' \deqn{g(\lambda_t) = \left(\beta_0 + \sum_{j = 1}^k \beta_j \cdot z_{t,j} \right),}
#' whereby \eqn{z_{t,j}} is the \eqn{j}-th covariate at time \eqn{t} and \eqn{g} is a link function. 
#' Currently supported link functions are: the identity \eqn{g(x) = x}, the logarithmic link
#' \eqn{g(x) = \ln x}, relu \eqn{g(x) = \max(x, \epsilon)} for small \eqn{\epsilon > 0}, and softplus
#' \eqn{g(x) = \log(\exp(x) - 1)}. To ensure positivity of \eqn{\lambda} if the identity function is used, \eqn{\beta_j, z_{t,j} > 0} must be enforced.
#' Alternatively, computational values of \eqn{\lambda \leq 0} can be set to a small positive value.
#' This option is named 'relu', due to its similarity to a ReLU function commonly used in machine learning.
#' The softplus link uses a numerically stable implementation of its inverse \eqn{\log(1 + \exp(x))}, ensuring \eqn{\lambda > 0} and smooth gradients everywhere.
#' 
#' 
#' Standard errors are computed by the square root of the diagonal elements of the inverse Hessian.
#' 
#' This function is implemented in two versions. The default runs on RCPP. 
#' An alternative version uses a \proglang{julia} implementation which can be chosen by 
#' setting the argument \proglang{julia} to TRUE. In order to
#' use this feature, a running \proglang{julia} installation is required on the system.
#' The RCPP implementation uses the derivative-free Nelder-Mead optimizer
#' to obtain parameter estimates. The \proglang{julia} implementation makes use of \proglang{julia}'s
#' automatic differentiation in order to obtain gradients such that it can use the LBFGS algorithm for
#' optimization. This enhances the numeric stability of the optimization and yields 
#' an internal validation if both methods yield qualitatively same parameter estimates.
#' Furthermore, the \proglang{julia} implementation can increase the computational speed
#' significantly, especially for large models. 
#' 
#' The model assessment tools \code{\link{cocoBoot}}, \code{\link{cocoPit}}, and \code{\link{cocoScore}}
#' will use a \proglang{julia} implementation as well, if the \code{\link{cocoReg}} was run with \proglang{julia}.
#' Additionally, one can make the RCPP output of \code{\link{cocoReg}} compatible with the \proglang{julia}
#' model assessments by setting \code{julia_installed} to true. In this case, the user can choose
#' between the \pkg{RCPP} and the \proglang{julia} implementation for model assessment.
#' 
#' @references 
#' Jung, R. C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 268--280.
#' 
#' Joe, H. (1996) Time series models with univariate margins in the convolution-closed infinitely divisible class. \emph{Journal of Applied Probability}, \bold{33}, 664--677.
#' @examples
#' ## GP2 model without covariates
#' length <- 1000
#' par <- c(0.5,0.2,0.05,0.3,0.3)
#' data <- cocoSim(order = 2, type = "GP", par = par, length = length)
#' fit <- cocoReg(order = 2, type = "GP", data = data)
#' 
#' ##Poisson1 model with covariates
#' length <- 1000
#' period <- 12
#' sin <- sin(2*pi/period*(1:length))
#' cos <- cos(2*pi/period*(1:length))
#' cov <- cbind(sin, cos)
#' par <- c(0.2, 0.2, -0.2)
#' data <- cocoSim(order = 1, type = "Poisson", par = par, xreg = cov, length = length)
#' fit <- cocoReg(order = 1, type = "Poisson", data = data, xreg = cov)
#' @export
cocoReg <- function(type, order, data, xreg = NULL, 
                    constrained.optim = TRUE, b.beta = -10,
                    start = NULL, start.val.adjust = TRUE, method_optim = "Nelder-Mead",
                    replace.start.val = 1e-5, iteration.start.val = 0.6,
                    method.hessian = "Richardson", cores=2, julia=FALSE, 
                    julia_installed=FALSE, link_function="log") {
  seasonality <- c(1, 2) #will be used as argument in future versions
  
  if ((type != "GP") & (type != "Poisson")) {
    stop("Option 'type' must be either Poisson or GP")
  }
  
  if (julia){
    start_time <- Sys.time()
    fit_julia <- cocoRegJulia(type, order, data, xreg, start, link_function, b.beta)
    end_time <- Sys.time()
    
    JuliaConnectoR::juliaLet("global fit2 = x", x=fit_julia)
    fit_no_optimization_in_dict <- juliaEval('
               delete!(fit2, "optimization")
               ')
    
    fit_R <- JuliaConnectoR::juliaGet(fit_no_optimization_in_dict)

    julia_out <- transformJuliaRegOutputToR(xreg=xreg, pars=fit_R[["values"]][[8]],
                                            grad=NULL, hes=NULL,
                                            inv_hes=fit_R[["values"]][[7]],
                                            se=fit_R[["values"]][[10]],
                                            data=data,
                                            type=type,
                                            order=order,
                                            likelihood=fit_R[["values"]][[1]],
                                            end_time = end_time,
                                            start_time=start_time, 
                                            julia_reg=fit_no_optimization_in_dict)

    class(julia_out) <- "coco"
    julia_out$link_function <- link_function

    return(julia_out)
  }
  
  if (is.null(xreg)) {
    
    output <- cocoReg_base(
      type = type, order = order, data = data, seasonality = seasonality, 
      constrained.optim = constrained.optim, start = start,
      start.val.adjust = start.val.adjust, replace.start.val = replace.start.val, method_optim=method_optim,
      iteration.start.val = iteration.start.val, method.hessian = method.hessian, julia_installed=julia_installed, link_function=link_function
    )
  } else {
    output <- cocoReg_cov(
      type = type, order = order, data = data, xreg = xreg, seasonality = seasonality,
      constrained.optim = constrained.optim, b.beta = b.beta, start = start, method_optim=method_optim,
      start.val.adjust = start.val.adjust, replace.start.val = replace.start.val,
      iteration.start.val = iteration.start.val, method.hessian = method.hessian, julia_installed=julia_installed, link_function=link_function
    )
  }
  
  class(output) <- "coco"
  
  return(output)
}

#' @exportS3Method
summary.coco <- function(object, ..., score = FALSE) {
  # Attach score flag to the object and set the class to "summary.coco"
  object$score <- score
  class(object) <- "summary.coco"
  return(object)
}

#' @exportS3Method
print.summary.coco <- function(x, ...) {
  coco <- x
  
  # Create data vectors with rounded values
  estimates <- round(coco$par, 4)
  se_values <- round(coco$se, 4)
  t_values <- round(coco$par / coco$se, 4)
  
  # Combine into a data frame for display
  df <- data.frame("Estimate" = estimates,
                   "Std. Error" = se_values,
                   "t" = t_values)
  
  # Generate parameter names based on model order and type
  if (coco$order == 1) {
    if (coco$type == "Poisson") {
      param_names <- c("lambda", "alpha")
    } else if (coco$type == "GP") {
      param_names <- c("lambda", "alpha", "eta")
    } else {
      param_names <- character(0)
    }
  } else if (coco$order == 2) {
    if (coco$type == "Poisson") {
      param_names <- c("lambda", "alpha1", "alpha2", "alpha3")
    } else if (coco$type == "GP") {
      param_names <- c("lambda", "alpha1", "alpha2", "alpha3", "eta")
    } else {
      param_names <- character(0)
    }
  } else {
    param_names <- character(0)
  }
  
  # Append covariate names if they exist
  if (!is.null(coco$cov)) {
    if (is.null(colnames(coco$cov))) {
      colnames(coco$cov) <- paste0("X", seq_len(ncol(coco$cov)))
    }
    # Remove "lambda" from the initial parameters and add covariate names
    param_names <- c(param_names[param_names != "lambda"], colnames(coco$cov))
  }
  
  # Set the row names of the coefficients dataframe
  rownames(df) <- param_names
  
  cat("Coefficients:\n")
  print(df, print.gap = 3, quote = FALSE, na.print = "")
  
  # Check if Julia regression flag is set
  julia <- !is.null(coco$julia_reg)
  
  # Build the output string with basic model statistics
  output_lines <- paste0(
    "\nType: ", coco$type,
    "\nOrder: ", coco$order,
    "\n\nLog-likelihood: ", round(coco$likelihood, 4)
  )
  
  if (isTRUE(coco$score)) {
    # When score flag is true, compute the additional scores
    score_vals <- cocoScore(coco, julia = julia)
    output_lines <- paste0(
      output_lines,
      "\nLogarithmic score: ", round(score_vals$log.score, 4),
      "\nQuadratic score: ", round(score_vals$quad.score, 4),
      "\nRanked probability score: ", round(score_vals$rps.score, 4),
      "\nAIC: ", round(score_vals$aic, 4),
      "\nBIC: ", round(score_vals$bic, 4)
    )
  } else {
    # Calculate AIC and BIC using the default formulas
    aic_value <- round(2 * length(coco$par) - 2 * coco$likelihood, 4)
    bic_value <- round(length(coco$par) * log(length(coco$ts)) - 2 * coco$likelihood, 4)
    output_lines <- paste0(
      output_lines,
      "\nAIC: ", aic_value,
      "\nBIC: ", bic_value
    )
  }
  
  cat(output_lines, "\n")
}

#' @exportS3Method
print.coco <- function(x, ...) {
  # Title and basic model information
  cat("Fitted coco Model\n")
  cat("====================================\n")
  cat("Type: ", x$type, "\n", sep = "")
  cat("Order: ", x$order, "\n", sep = "")
  cat("Link Function: ", x$link_function, "\n", sep = "")
  
  # Coefficient information (with names assigned via coef.coco)
  cat("\nCoefficients:\n")
  print(coef(x))
  
  # Log-likelihood
  cat("\nLog-likelihood: ", round(x$likelihood, 4), "\n", sep = "")
  
  # Duration of model estimation (if available)
  if(!is.null(x$duration)) {
    cat("Duration (seconds): ", as.numeric(x$duration), "\n", sep = "")
  }
  
  # Covariate information (if available)
  if(!is.null(x$cov)) {
    cov_names <- colnames(x$cov)
    if(is.null(cov_names)) {
      cov_names <- paste0("X", seq_len(ncol(x$cov)))
    }
    cat("Covariates: ", paste(cov_names, collapse = ", "), "\n")
  }
  
  invisible(x)
}

#' @exportS3Method
fitted.coco <- function(object, ...) {
  # Compute the residuals (fitted values, residuals, and other diagnostics)
  res <- cocoResid(object, ...)
  # Return only the fitted values
  res$fitted
}

#' @exportS3Method
residuals.coco <- function(object, ...) {
  res <- cocoResid(object, ...)
  res$residuals
}

#' @exportS3Method
coef.coco <- function(object, ...) {
  params <- object$par
  names(params) <- get_coco_param_names(object)
  params
}


#' @exportS3Method
vcov.coco <- function(object, ...) {
  if (!is.null(object[["inv hessian"]])) {
    V <- object[["inv hessian"]]
  } else {
    stop("Covariance matrix not available for this object")
  }
  
  param_names <- get_coco_param_names(object)
  rownames(V) <- param_names
  colnames(V) <- param_names
  V
}

#' @exportS3Method
rstandard.coco <- function(model, ...) {
  res <- cocoResid(model, ...)
  res$pe.resid
}

#' @exportS3Method
nobs.coco <- function(object, ...) {
  length(object$ts)
}

#' @exportS3Method
extractAIC.coco <- function(fit, scale = 0, k = 2, ...) {
  df <- length(fit$par)
  aic_value <- -2 * fit$likelihood + k * df
  c(df, aic_value)
}

#' @exportS3Method
logLik.coco <- function(object, ...) {
  val <- object$likelihood
  npar <- length(object$par)
  
  attr(val, "nobs") <- length(object$ts)
  attr(val, "df") <- npar
  class(val) <- "logLik"
  val
}

get_coco_param_names <- function(object) {
  # Define parameter names based on the model order and type.
  if (object$order == 1) {
    if (object$type == "Poisson") {
      param_names <- c("lambda", "alpha")
    } else if (object$type == "GP") {
      param_names <- c("lambda", "alpha", "eta")
    } else {
      param_names <- character(0)
    }
  } else if (object$order == 2) {
    if (object$type == "Poisson") {
      param_names <- c("lambda", "alpha1", "alpha2", "alpha3")
    } else if (object$type == "GP") {
      param_names <- c("lambda", "alpha1", "alpha2", "alpha3", "eta")
    } else {
      param_names <- character(0)
    }
  } else {
    param_names <- character(0)
  }
  
  # If covariates are included, remove "lambda" and append the covariate names.
  if (!is.null(object$cov)) {
    if (is.null(colnames(object$cov))) {
      colnames(object$cov) <- paste0("X", seq_len(ncol(object$cov)))
    }
    param_names <- c(param_names[param_names != "lambda"], colnames(object$cov))
  }
  
  param_names
}

#' @exportS3Method
autoplot.coco <- function(object, which = "fitted", ...) {
  # Ensure ggplot2 namespace is loaded
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for this function to work")
  }
  
  # Calculate fitted values and residuals using your internal function.
  resid_out <- cocoResid(object, ...)
  ts_obs <- object$ts
  # Generate a time index; note that fitted values and residuals are computed for observations
  # starting at index (order + 1)
  time_index <- seq_along(ts_obs)
  valid_index <- time_index[(object$order + 1):length(ts_obs)]
  
  if (which == "fitted") {
    # Create a data frame for plotting the observed and fitted values.
    df <- data.frame(Time = valid_index,
                     Observed = ts_obs[(object$order + 1):length(ts_obs)],
                     Fitted = resid_out$fitted)
    
    p <- ggplot2::ggplot(df, ggplot2::aes(x = Time)) +
      ggplot2::geom_line(ggplot2::aes(y = Observed), color = "black") +
      ggplot2::geom_line(ggplot2::aes(y = Fitted), color = "steelblue", linetype = "dashed") +
      ggplot2::labs(title = "Observed vs. Fitted Values",
                    y = "Count") +
      ggplot2::theme_bw()
    
    return(p)
    
  } else if (which == "residuals") {
    # Plot the standardized (Pearson) residuals over time.
    df <- data.frame(Time = valid_index,
                     Residuals = resid_out$pe.resid)
    
    p <- ggplot2::ggplot(df, ggplot2::aes(x = Time, y = Residuals)) +
      ggplot2::geom_line(color = "black") +
      ggplot2::labs(title = "Pearson Residuals",
                    y = "Pearson Residual") +
      ggplot2::theme_bw()
    
    return(p)
    
  } else if (which == "acf") {
    # Create an ACF plot of the standardized residuals.
    acf_obj <- stats::acf(resid_out$pe.resid, plot = FALSE)
    df <- data.frame(Lag = as.numeric(acf_obj$lag),
                     ACF = acf_obj$acf)
    
    p <- ggplot2::ggplot(df, ggplot2::aes(x = Lag, y = ACF)) +
      ggplot2::geom_hline(yintercept = 0, linetype = "dashed") +
      ggplot2::geom_bar(stat = "identity", fill = "steelblue") +
      ggplot2::labs(title = "ACF of Pearson Residuals") +
      ggplot2::theme_bw()
    
    return(p)
    
  } else {
    stop("Invalid 'which' argument. Choose from 'fitted', 'residuals', or 'acf'.")
  }
}

#' @exportS3Method
plot.coco <- function(x, interactive = TRUE, ...) {
  
  if (interactive) {
    # Display the standardized residuals plot first.
    p_resid <- autoplot.coco(x, which = "residuals", ...)
    print(p_resid)
    
    # Wait for user input before continuing.
    readline(prompt = "Press Enter to view the ACF plot...")
    
    # Now display the ACF plot of the standardized residuals.
    p_acf <- autoplot.coco(x, which = "acf", ...)
    print(p_acf)
    
  } else {
    # For non-interactive usage, default to a single plot.
    # For example, the observed vs. fitted plot.
    p <- autoplot.coco(x, which = "acf", ...)
    print(p)
  }
}

#-----------------cgeck install julia packages----
test_that("insatll_Julia_packages", {
  skip_on_cran()
  expect_no_error(installJuliaPackages())
})
#-----------------No covariates, constrained--------------------------------
test_that("GP2Works", {
  skip_on_cran()
  length <- 300
  par <- c(0.5,0.2,0.05,0.3,0.3) #lambda, alpha1, alpha2, alpha3, eta
  set.seed(123499)
  data <- cocoSim(order = 2, type = "GP", par = par, length = length)
  cocoSim(order = 2, type = "GP", par = par, length = length, julia = TRUE,
          julia_seed = 1234992)
  
  fit <- cocoReg(order = 2, type = "GP", data = data, julia_installed = TRUE)
  for (i in c(TRUE, FALSE)){
    cocoPit(fit, julia=i)
    cocoResid(fit)
    summary(fit, julia=i)
    cocoScore(fit, julia=i)
    predict(fit, k=2, julia=i)
    cocoBoot(fit, rep.Bootstrap = 6, julia=i)
  }
  expect_no_error(fit)
})

test_that("Poisson2Works", {
  skip_on_cran()
  length <- 300
  par <- c(0.5, 0.2, 0.05, 0.3)
  set.seed(12347)
  data <- cocoSim(order = 2, type = "Poisson", par = par, length = length)
  cocoSim(order = 2, type = "Poisson", par = par, length = length, 
                      julia = TRUE,
                      julia_seed = 12342399)
  
  fit <- cocoReg(order = 2, type = "Poisson", data = data, julia_installed = TRUE)
  for (i in c(TRUE, FALSE)){
    cocoPit(fit, julia=i)
    cocoResid(fit)
    summary(fit, julia=i)
    cocoScore(fit, julia=i)
    predict(fit, k=3, julia=i)
    cocoBoot(fit, rep.Bootstrap = 6, julia=i)
  }
  expect_no_error(fit)
})

test_that("GP1Works", {
  skip_on_cran()
  length <- 300
  par <- c(0.5, 0.2, 0.2)
  set.seed(12341)
  data <- cocoSim(order = 1, type = "GP", par = par, length = length)
  cocoSim(order = 1, type = "GP", par = par, length = length, 
          julia = TRUE,
          julia_seed = 123423299)
  
  fit <- cocoReg(order = 1, type = "GP", data = data, julia_installed = TRUE)
  for (i in c(TRUE, FALSE)){
    cocoPit(fit, julia=i)
    cocoResid(fit)
    summary(fit, julia=i)
    cocoScore(fit, julia=i)
    predict(fit, k=4, julia=i)
    cocoBoot(fit, rep.Bootstrap = 6, julia=i)
  }
  expect_no_error(fit)
})

test_that("Poisson1Works", {
  skip_on_cran()
  length <- 300
  par <- c(0.5, 0.2)
  set.seed(12345)
  data <- cocoSim(order = 1, type = "Poisson", par = par, length = length)
  
  cocoSim(order = 1, type = "Poisson", par = par, length = length, 
          julia = TRUE,
          julia_seed = 1234232199)
  fit <- cocoReg(order = 1, type = "Poisson", data = data, julia_installed = TRUE)
  for (i in c(TRUE, FALSE)){
    cocoPit(fit, julia=i)
    cocoResid(fit)
    summary(fit, julia=i)
    cocoScore(fit, julia=i)
    predict(fit, k=4, julia=i)
    cocoBoot(fit, rep.Bootstrap = 6, julia=i)
  }
  expect_no_error(fit)
})

#-----------------Covariates, constrained--------------------------------
test_that("Poisson1Works_cov", {
  skip_on_cran()
 ##Poisson1 model with covariates
 length <- 300
 period <- 50
 sin <- sin(2*pi/period*(1:length))
 cos <- cos(2*pi/period*(1:length))
 cov <- cbind(sin, cos)
 par <- c(0.8, 0.2, -0.2)
 set.seed(1234)
 data <- cocoSim(order = 1, type = "Poisson", par = par,
                     xreg = cov, length = length)
 
 cocoSim(order = 1, type = "Poisson", par = par, xreg = cov, length = length, 
         julia = TRUE,
         julia_seed = 123499)
 fit <- cocoReg(order = 1, type = "Poisson", data = data, xreg = cov, 
                julia_installed = TRUE)
 for (i in c(TRUE, FALSE)){
   cocoPit(fit, julia=i)
   cocoResid(fit)
   summary(fit, julia=i)
   cocoScore(fit, julia=i)
   predict(fit, k=2, xcast = cov[1:2,], julia=i)
   cocoBoot(fit, rep.Bootstrap = 6, julia=i)
 }
 expect_no_error(fit)

})

test_that("GP1Works_cov", {
  skip_on_cran()
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.3, 0.2, 0.2, -0.2)
  set.seed(1234567)
  data <- cocoSim(order = 1, type = "GP", par = par,
                      xreg = cov, length = length)
  
  cocoSim(order = 1, type = "GP", par = par, xreg = cov, length = length, 
          julia = TRUE,
          julia_seed = 123499)
  fit <- cocoReg(order = 1, type = "GP", data = data, xreg = cov, julia_installed = TRUE)
  for (i in c(TRUE, FALSE)){
    cocoPit(fit, julia=i)
    cocoResid(fit)
    summary(fit, julia=i)
    cocoScore(fit, julia=i)
    predict(fit, k=2, xcast = cov[1:2,], julia=i)
    cocoBoot(fit, rep.Bootstrap = 6, julia=i)
  }
  expect_no_error(fit)
})

test_that("Poisson2Works_cov", {
  skip_on_cran()
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.2, 0.05, 0.3, 0.2, -0.2)
  set.seed(12345678)
  data <- cocoSim(order = 2, type = "Poisson", par = par,
                      xreg = cov, length = length)
  
  cocoSim(order = 2, type = "Poisson", par = par, xreg = cov, length = length, 
          julia = TRUE,
          julia_seed = 123499)
  fit <- cocoReg(order = 2, type = "Poisson", data = data, xreg = cov, julia_installed = TRUE)
  for (i in c(TRUE, FALSE)){
    cocoPit(fit, julia=i)
    cocoResid(fit)
    summary(fit, julia=i)
    cocoScore(fit, julia=i)
    predict(fit, k=2, xcast = cov[1:2,], julia=i)
    cocoBoot(fit, rep.Bootstrap = 6, julia=i)
  }
  expect_no_error(fit)
})

test_that("GP2Works_cov", {
  skip_on_cran()
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.2, 0.05, 0.3, 0.2, 0.2, -0.2)
  set.seed(123456798)
  data <- cocoSim(order = 2, type = "GP", par = par,
                      xreg = cov, length = length)
  
  cocoSim(order = 2, type = "GP", par = par, xreg = cov, length = length, 
          julia = TRUE,
          julia_seed = 123499)
  fit <- cocoReg(order = 2, type = "GP", data = data, xreg = cov, julia_installed = TRUE)
  for (i in c(TRUE, FALSE)){
    cocoPit(fit, julia=i)
    cocoResid(fit)
    summary(fit, julia=i)
    cocoScore(fit, julia=i)
    predict(fit, k=2, xcast = cov[1:2,], julia=i)
    cocoBoot(fit, rep.Bootstrap = 6, julia=i)
  }
  expect_no_error(fit)
})

#-----------------No covariates, unconstrained--------------------------------
test_that("GP2Works_constr", {
  length <- 300
  par <- c(0.5,0.2,0.05,0.3,0.3)
  set.seed(123499)
  data <- cocoSim(order = 2, type = "GP", par = par, length = length)
  
  fit <- cocoReg(order = 2, type = "GP", data = data, constrained.optim = FALSE)
  expect_no_error(fit)
})

test_that("Poisson2Works_constr", {
  length <- 300
  par <- c(0.5, 0.2, 0.05, 0.3)
  set.seed(12347)
  data <- cocoSim(order = 2, type = "Poisson", par = par, length = length)
  
  fit <- cocoReg(order = 2, type = "Poisson", data = data, constrained.optim = FALSE)
  expect_no_error(fit)
})

test_that("GP1Works_constr", {
  length <- 300
  par <- c(0.5, 0.2, 0.2)
  set.seed(12341)
  data <- cocoSim(order = 1, type = "GP", par = par, length = length)
  
  fit <- cocoReg(order = 1, type = "GP", data = data, constrained.optim = FALSE)
  expect_no_error(fit)
})

test_that("Poisson1Works_constr", {
  length <- 300
  par <- c(0.5, 0.2)
  set.seed(12345)
  data <- cocoSim(order = 1, type = "Poisson", par = par, length = length)
  
  fit <- cocoReg(order = 1, type = "Poisson", data = data, constrained.optim = FALSE)
  expect_no_error(fit)
})

#-----------------Covariates, unconstrained--------------------------------
test_that("Poisson1Works_cov_constr", {
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.8, 0.2, -0.2)
  set.seed(1234)
  data <- cocoSim(order = 1, type = "Poisson", par = par,
                      xreg = cov, length = length)
  
  fit <- cocoReg(order = 1, type = "Poisson", data = data, xreg = cov, constrained.optim = FALSE)
  expect_no_error(fit)
})

test_that("GP1Works_cov_constr", {
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.8, 0.2, 0.2, -0.2)
  set.seed(1234567)
  data <- cocoSim(order = 1, type = "GP", par = par,
                      xreg = cov, length = length)
  
  fit <- cocoReg(order = 1, type = "GP", data = data, xreg = cov, constrained.optim = FALSE)
  expect_no_error(fit)
})

test_that("GP2Works_cov_uncosntrained", {
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.2, 0.05, 0.3, 0.2, 0.2, -0.2)
  set.seed(123456798)
  data <- cocoSim(order = 2, type = "GP", par = par,
                      xreg = cov, length = length)
  
  fit <- cocoReg(order = 2, type = "GP", data = data, xreg = cov,
                 constrained.optim = FALSE)
  expect_no_error(fit)
})

test_that("Poisson2Works_cov_constr", {
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.2, 0.05, 0.3, 0.2, -0.2)
  set.seed(12345678)
  data <- cocoSim(order = 2, type = "Poisson", par = par,
                      xreg = cov, length = length)
  
  fit <- cocoReg(order = 2, type = "Poisson", data = data, xreg = cov, constrained.optim = FALSE)
  expect_no_error(fit)
})



#-----------------Simulate initial time series-------------
test_that("initialize_time_series", {
  length <- 300
  par <- c(0.5,0.2,0.05,0.3,0.3)
  set.seed(123499)
  data <- cocoSim(order = 2, type = "GP", par = par, length = length)
  
  data2 <- cocoSim(order = 2, type = "GP", par = par, length = length,
                      init = data)
  expect_no_error(data)
})
#----------------Custom starting values covariates------------------------------
test_that("custom_start_values_GP2", {
  length <- 300
  par <- c(0.5,0.2,0.05,0.3,0.3)
  set.seed(122133499)
  data <- cocoSim(order = 2, type = "GP", par = par, length = length)
  
  fit <- cocoReg(order = 2, type = "GP", data = data, start = par)
  expect_no_error(fit)
})

test_that("custom_start_values_Poisson2", {
  length <- 300
  par <- c(0.5, 0.2, 0.05, 0.3)
  set.seed(13423499)
  data <- cocoSim(order = 2, type = "Poisson", par = par, length = length)
  
  fit <- cocoReg(order = 2, type = "Poisson", data = data, start = par)
  expect_no_error(fit)
})

test_that("custom_start_values_GP1", {
  length <- 300
  par <- c(0.5,0.2,0.3)
  set.seed(12323499)
  data <- cocoSim(order = 1, type = "GP", par = par, length = length)
  
  fit <- cocoReg(order = 1, type = "GP", data = data, start = par)
  expect_no_error(fit)
})

test_that("custom_start_values_Poisson1", {
  length <- 300
  par <- c(0.5,0.2)
  set.seed(12349923)
  data <- cocoSim(order = 1, type = "Poisson", par = par, length = length)
  
  fit <- cocoReg(order = 1, type = "Poisson", data = data, start = par)
  expect_no_error(fit)
})
#----------------Custom starting values covariates------------------------------


test_that("GP1Works_cov_2", {
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.8, 0.2, 0.2, -0.2)
  set.seed(1234567)
  data <- cocoSim(order = 1, type = "GP", par = par,
                      xreg = cov, length = length)
  
  fit <- cocoReg(order = 1, type = "GP", data = data, xreg = cov)
  cocoPit(fit)
  cocoResid(fit)
  summary(fit)
  cocoScore(fit)
  predict(fit, k=2, xcast = cov[1:2,])
  cocoBoot(fit, rep.Bootstrap = 6)
  expect_no_error(fit)
})

test_that("Poisson2Works_cov_2", {
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.2, 0.05, 0.3, 0.2, -0.2)
  set.seed(12345678)
  data <- cocoSim(order = 2, type = "Poisson", par = par,
                      xreg = cov, length = length)
  
  fit <- cocoReg(order = 2, type = "Poisson", data = data, xreg = cov)
  cocoPit(fit)
  cocoResid(fit)
  summary(fit)
  cocoScore(fit)
  predict(fit, k=2, xcast = cov[1:2,])
  cocoBoot(fit, rep.Bootstrap = 6)
  expect_no_error(fit)
})

test_that("GP2Works_cov_2", {
  ##Poisson1 model with covariates
  length <- 300
  period <- 50
  sin <- sin(2*pi/period*(1:length))
  cos <- cos(2*pi/period*(1:length))
  cov <- cbind(sin, cos)
  par <- c(0.2, 0.05, 0.3, 0.2, 0.2, -0.2)
  set.seed(123456798)
  data <- cocoSim(order = 2, type = "GP", par = par,
                      xreg = cov, length = length)
  
  fit <- cocoReg(order = 2, type = "GP", data = data, xreg = cov)
  cocoPit(fit)
  cocoResid(fit)
  summary(fit)
  cocoScore(fit)
  predict(fit, k=2, xcast = cov[1:2,])
  cocoBoot(fit, rep.Bootstrap = 6)
  expect_no_error(fit)
})
#----------------Wrong models no covariates-----------------------------------
test_that("wrong_modelsGP2", {
  skip_on_cran()
  length <- 300
  par <- c(0.5,0.2,0.05,0.3,0.3)
  set.seed(123499)
  data <- cocoSim(order = 2, type = "GP", par = par, length = length)
  
  for (julia in c(TRUE, FALSE)){
    fit <- cocoReg(order = 1, type = "Poisson", data = data, julia = julia)
    fit <- cocoReg(order = 2, type = "Poisson", data = data, julia = julia)
    fit <- cocoReg(order = 1, type = "GP", data = data, julia = julia)
  }
  expect_no_error(fit)
})

test_that("wrong_modelsPoisson2", {
  skip_on_cran()
  length <- 300
  par <- c(0.5, 0.2, 0.05, 0.3)
  set.seed(12347)
  data <- cocoSim(order = 2, type = "Poisson", par = par, length = length)
  
  for (julia in c(TRUE, FALSE)){
    fit <- cocoReg(order = 1, type = "Poisson", data = data,  julia = julia)
    fit <- cocoReg(order = 1, type = "GP", data = data,  julia = julia)
    fit <- cocoReg(order = 2, type = "GP", data = data,  julia = julia)
  }
  expect_no_error(fit)
  
})

test_that("wrong_modelsGP1", {
  skip_on_cran()
  length <- 300
  par <- c(0.5, 0.2, 0.2)
  set.seed(12341)
  data <- cocoSim(order = 1, type = "GP", par = par, length = length)
  
  for (julia in c(TRUE, FALSE)){
    fit <- cocoReg(order = 1, type = "Poisson", data = data,  julia = julia)
    fit <- cocoReg(order = 2, type = "Poisson", data = data,  julia = julia)
    fit <- cocoReg(order = 2, type = "GP", data = data,  julia = julia)
  }
  expect_no_error(fit)
})

test_that("wrong_modelsPoisson1", {
  skip_on_cran()
  length <- 300
  par <- c(0.5, 0.2)
  set.seed(12345)
  data <- cocoSim(order = 1, type = "Poisson", par = par, length = length)
  
  for (julia in c(TRUE, FALSE)){
    fit <- cocoReg(order = 2, type = "Poisson", data = data,  julia = julia)
    fit <- cocoReg(order = 1, type = "GP", data = data,  julia = julia)
    fit <- cocoReg(order = 2, type = "GP", data = data,  julia = julia)
  }
  expect_no_error(fit)

})

#' @title Simulation of Count Time Series
#' @description The function generates a time series of low counts from the (G)PAR model class for a specified
#'  innovation distribution, sample size, lag order,
#' and parameter values. 
#' @param type character, either "Poisson" or "GP" indicating the type of the innovation distribution
#' @param order integer, either 1 or 2 indicating the order of the model
#' @param par numeric vector, the parameters of the model, the number of elements
#'  in the vector depends on the type and order specified. 
#' @param length integer, the number of observations in the generated time series
#' @param xreg data frame of control variables (defaul: NULL)
#' @param init numeric vector, initial data to use (default: NULL). See details
#' for more information on the usage.
#' @param julia If TRUE, the \proglang{julia} implementation is used. In this case, init is ignored but it might be faster (default: FALSE).
#' @param link_function Specifies the link function for the conditional mean of the innovation (\eqn{\lambda}). The default is `log`, but other available options include `identity` and `relu`. This parameter is applicable only when covariates are used. Note that using the `identity` link function may result in \eqn{\lambda} becoming negative. To prevent this, ensure all covariates are positive and restrict the parameter \eqn{\beta} to positive values.
#' @param julia_seed Seed for the \proglang{julia} implementation. Only used if \proglang{julia} equals TRUE.
#' @return a vector of the simulated time series 
#' @details The function checks for valid input of the type, order, parameters, and initial data
#' before generating the time series.
#'
#'The init parameter allows users to set a custom burn-in period
#'for the simulation. By default, when simulating with covariates, no burn-in
#'period is specified since there is no clear choice on the covariates.
#'However, the init argument gives users the flexibility to select an
#'appropriate burn-in period for the covariate case. One way to do this is to
#'simulate a time series using \code{\link{cocoSim}} with appropriate covariates and pass the
#'resulting time series to the
#'init argument of a new \code{\link{cocoSim}} run so that the first time series is used as
#'the burn-in period. 
#'If init is not specified for the covariate case, a warning will be returned
#'to prompt the user to specify a custom burn-in period. This helps ensure that
#'the simulation accurately captures the dynamics of the system being modeled.
#' @author Manuel Huth
#' @examples
#' #First Order Model
#' lambda <- 1
#' alpha <- 0.4
#' set.seed(12345)
#' 
#' # Simulate using the RCPP implementation
#' data_rcpp <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)
#' 
#' @examples
#' # Second Order Model
#' lambda <- 1
#' alpha_1 <- 0.3
#' alpha_2 <- 0.1
#' alpha_3 <- 0.2
#' eta <- 0.2
#' data_j <- cocoSim(order = 2, type = "GP",
#'                   par = c(lambda, alpha_1, alpha_2,alpha_3,eta),
#'                   length = 100)
#' @export

cocoSim <- function(type, order, par, length, xreg = NULL, init = NULL, 
                    julia=FALSE, julia_seed = NULL, link_function="log") {
  seasonality <- c(1, 2) #will be used as argument in future versions
  
  
  
  if (is.null(init)) {
    length_burn_in <- 200
    init_add <- 0
  } else {
    length_burn_in <- 0
    init_add <- order
  }
  
  if (is.null(xreg)) {
    if (order == 2){
      
      if (par[2] + par[4] > 1) {
        stop("The condition alpha_1 + alpha_3 < 1 ist not satisfied")
      }
      
      if ((par[2] < 0) | (par[3] < 0) | (par[4] < 0) | (par[2] >= 1) | (par[3] >= 1) | (par[4] >= 1)) {
        stop("The alpha parameters must be within the unit interval")
      }
      

      if ((type == "GP") & ((par[5] < 0) | (par[5] >= 1))) {
          stop("eta must be within the unit interval")
      }


    } else {
      if ((par[2] < 0) | (par[2] >= 1))  {
        stop("alpha must be within the unit interval")
      }
      if ((type == "GP") & ((par[3] < 0) | (par[3] >= 1))){
        stop("eta must be within the unit interval")
      }
    }
    
    if (julia){
      if (!is.null(julia_seed)){
        setJuliaSeed(julia_seed)
      }
      return(cocoSimJulia(type, order, par, length, xreg, link_function))
    }
    
    size <- length + length_burn_in + init_add
    output <- cocoSim_base(type = type, order = order, par = par, size = size,
                           seasonality = seasonality, init = init)
    output$data <- output$data[(length_burn_in+1):(length+length_burn_in)]
  } else {
    if (order == 2){
      
      if (par[1] + par[3] > 1) {
        stop("The condition alpha_1 + alpha_3 < 1 ist not satisfied")
      }
      
      if ((par[1] < 0) | (par[2] < 0) | (par[3] < 0) | (par[1] >= 1) | (par[2] >= 1) | (par[3] >= 1)) {
        stop("The alpha parameters must be within the unit interval")
      }

      if ((type == "GP") & ((par[4] < 0) | (par[4] >= 1))){
        stop("eta must be within the unit interval")
      }
      
    } else {
      if ((par[1] < 0) | (par[1] >= 1))  {
        stop("alpha must be within the unit interval")
      }
      if ((type == "GP") & ((par[2] < 0) | (par[2] >= 1))){
        stop("eta must be within the unit interval")
      }
    }
    
    if (julia){
      setJuliaSeed(julia_seed)
      return(cocoSimJulia(type, order, par, length, xreg, link_function))
    }
    size <- length + init_add
    output <- cocoSim_cov(
      type = type, order = order, par = par, size = size, xreg = xreg,
      seasonality = seasonality, init = init, link_function=link_function
    )
    
  }

  return(output$data)
}

#' @importFrom JuliaConnectoR juliaCall
cocoSimJulia <- function(type, order, par, length, xreg, link_function){
  addJuliaFunctions()
  
  if (is.null(xreg)){
    par_use <- c(par[2:length(par)], par[1])
  } else {
    par_use <- par
  }
  
  return(JuliaConnectoR::juliaCall("cocoSim", type, order, par_use, length, xreg, link_function))
}

cocoForecastKStepsRCPP <- function(fit, k=1, number_simulations=500, covariates=NULL){
  
  init <- fit$ts[length(fit$ts)]
  if (fit$order == 2){
    init <- fit$ts[(length(fit$ts)-1):length(fit$ts)]
  }
  
  run_sim <- function(i){ 
    #cannot be cocosim as the init argument is not available in cocoSim
    if (!is.null(covariates)){
      output <- cocoSim_cov(fit$type, order = fit$order, par = fit$par, 
                          size= k+length(init), xreg = covariates, 
                          seasonality = c(1, 2), init = init, link_function=fit$link_function)$data
    } else {
      output <- cocoSim_base(fit$type, order = fit$order, par = fit$par, 
                            size= k+length(init),
                            seasonality = c(1, 2), init = init)$data
    }
  }
  
  if (k > 1) {
    out_matrix <- t(sapply(1:number_simulations, run_sim))
  } else if (k==1){
    out_matrix <- matrix(sapply(1:number_simulations, run_sim), ncol=1)
  }
  
  freq_table <- function(i){
    df <- as.data.frame(table(out_matrix[,i]) / number_simulations)
    colnames <- c("value", "frequency")
    colnames(df) <- colnames
    df["value"] <- as.numeric(levels(df[, "value"]))

    return(df)
  }
  
  return(lapply(1:ncol(out_matrix), freq_table))
  
}

#' @importFrom JuliaConnectoR juliaEval
addJuliaFunctions <- function(){
  
  if (JuliaConnectoR::juliaEval("isdefined(Main, :cocoReg)")) {
  } else {
    #----------------Add Julia functions-----------------------------------
    JuliaConnectoR::juliaEval('
      using Coconots
      using Random
    ')
    #-------------------------------------
  }
}

#' @title Residual Based Model Assessment Procedure
#' @description Calculates the (Pearson) residuals of a fitted model for model evaluation purposes.
#' @param coco An object of class "coco
#' @param val.num A non-negative real number that halts the calculation once the cumulative probability reaches 1-\code{val.num}
#' @author Manuel Huth
#' @importFrom stats var fitted residuals
#' @importFrom ggplot2 autoplot
#' @return a list that includes the (Pearson) residuals, conditional expectations, conditional variances,
#' and information on the model specifications.
#'@details The Pearson residuals are computed as the scaled
#'deviation of the observed count from its conditional expectation given the relevant 
#'past history, including covariates, if applicable. If a fitted model is correctly specified,
#'the Pearson residuals should exhibit mean zero, variance one, and no significant serial correlation. 
#'@export
cocoResid <- function(coco, val.num = 1e-10) {
  start.time <- Sys.time()

  if (val.num <= 0) {
    stop("Option val.num must be a non-negative real number")
  }

  seasonality <- c(1, 2) #will be used as argument in future versions coco$seasonality
  bool_covariates <- !is.null(coco$cov)
  par <- coco$par
  data <- coco$ts
  
  #handle covariates
  if (bool_covariates){
    # set up values for lambda
    xreg <- coco$cov
    vec_lambda <- par[(length(par)-ncol(xreg)+1):length(par)]
    lambdas <- c()
    for (j in 1:length(data)) {
      lambdas[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
    }
  } else {
    lambda <- par[1]
  }
    

  ### GP/PAR1
  if (coco$order == 1) {

    alpha <- par[2-bool_covariates]
      
    if (coco$type == "Poisson"){
      eta <- 0
    } else {
      eta <- par[3-bool_covariates]
    }
      
    # Fitted Values
    fitted <- c(rep(mean(data), seasonality[1])) 
    varX <- c(rep(stats::var(data), seasonality[1])) 
      
    fitted_varX <- lapply((seasonality[1] + 1):(length(data)), function(j) {
      if (bool_covariates){
        lambda <- lambdas[j]
      }
      fitted_value <- alpha * data[j - seasonality[1]] + lambda / (1-eta)
      varX_value <- alpha * (1 - alpha) * data[j - seasonality[1]] + lambda / (1-eta)^3
      return(list(fitted = fitted_value, varX = varX_value))
    })
      
    fitted <- unlist(lapply(fitted_varX, function(x) x$fitted))
    varX <- unlist(lapply(fitted_varX, function(x) x$varX))
      
  } # end if GP/PAR1


  ### GP/PAR2
  if (coco$order == 2) {
      
    alpha1 <- par[2-bool_covariates]
    alpha2 <- par[3-bool_covariates]
    alpha3 <- par[4-bool_covariates]
      
    U <- 1 / (1 - alpha1 - alpha2 - alpha3)
      
    if (coco$type == "Poisson"){
      eta <- 0
    } else {
      eta <- par[5-bool_covariates]
    }
      
    # Fitted Values
    fitted <- c(rep(mean(data), seasonality[2])) 
    varX <- c(rep(stats::var(data), seasonality[2])) 
      
    fitted_varX <- lapply((seasonality[2] + 1):(length(data)), function(j) {
      if (bool_covariates){
        lambda <- lambdas[j]
      }
        
      y <- data[j - seasonality[1]]
      z <- data[j - seasonality[2]]
        
      # compute conditional mean of R
      meanR <- 0
      varR <- 0
      prob_accumulated <- 0
      probabilities <- c()
      r <- 0
      while (prob_accumulated < (1 - val.num)) {
        prob <- dR2(r, y, z, lambda, alpha1, alpha2, alpha3, eta)
        probabilities <- c(probabilities, prob)
        prob_accumulated <- prob_accumulated + prob
        r <- r + 1
      }
      
      #define objects for cond. mean and cond. variance computation
      n_relevant_probabilities <- r-1 #r-1 as +1 is added in the end within the while loop
      relevant_counts <- 0:(r-1)
      
      #compute conditional mean and conditional variance
      meanR <- sum(probabilities * relevant_counts)
      varR <-  sum(probabilities * ((relevant_counts - meanR)^2))
        
      fitted_value <- meanR + lambda / (1 - eta)
      varX_value <- varR + lambda / (1 - eta)^3

      return(list(fitted = fitted_value, varX = varX_value))
    })
      
    fitted <- unlist(lapply(fitted_varX, function(x) x$fitted))
    varX <- unlist(lapply(fitted_varX, function(x) x$varX))
      
  } # end if GP/PAR2


  # Residuals
  residuals <- data[(coco$order+1):length(data)] - fitted

  # Pearson Residuals
  peResid <- residuals / varX^0.5

  end.time <- Sys.time()

  time <- end.time - start.time

  list_out <- list(
    "fitted" = fitted, "residuals" = residuals, "pe.resid" = peResid,
    "cond.var" = varX, "type" = coco$type, "order" = coco$order, "ts" = coco$ts,
    "par" = par,"duration" = time, xreg = coco$cov
  )

  class(list_out) <- "cocoResid"

  return(list_out)
} # end function


#------------------S3 methods--------------------------------------------------
#' @importFrom ggplot2 autoplot theme_bw ggtitle
#' @export
ggplot2::autoplot
#' @exportS3Method
autoplot.cocoResid <- function(object, ...){
  forecast::ggAcf(object$pe.resid) + theme_bw() + ggtitle("Pearson Residuals") + ggplot2::theme(text = ggplot2::element_text(size = 20)) 
}

#' @exportS3Method
plot.cocoResid <- function(x, ...) {
  p <- autoplot(
    x,
    ...
  )
  print(p)
}

#'@exportS3Method
print.cocoResid <- function(x, ...) {
  print(autoplot(x, ...,))
  invisible(x)
}

#' @exportS3Method
print.cocoResid <- function(x, ...) {
  cat("cocoResid Object\n")
  cat("Model Type: ", x$type, "\n")
  cat("Model Order: ", x$order, "\n")
  cat("Number of Fitted Values: ", length(x$fitted), "\n")
  
  # Display covariate information
  if (is.null(x$xreg)) {
    cat("Covariates: None\n")
  } else {
    n_cov <- if (is.data.frame(x$xreg) || is.matrix(x$xreg)) ncol(as.matrix(x$xreg)) else 1
    cat("Covariates: Present (", n_cov, " variable(s))\n", sep = "")
  }
  
  cat("Calculation Duration: ", as.numeric(x$duration, units = "secs"), " seconds\n", sep = "")
  cat("\n")
  
  invisible(x)
}

#' @exportS3Method
summary.cocoResid <- function(object, ...) {
  # Build a summary list extracting key information from the cocoResid object
  res <- list(
    model_type = object$type,
    model_order = object$order,
    ts_length = length(object$ts),
    n_fitted = length(object$fitted),
    covariates = if (is.null(object$xreg)) {
      "None"
    } else if (is.data.frame(object$xreg) || is.matrix(object$xreg)) {
      paste("Present (", ncol(as.matrix(object$xreg)), " variable(s))", sep = "")
    } else {
      "Present (1 variable)"
    },
    residuals_summary = summary(object$residuals),
    pearson_residuals_summary = summary(object$pe.resid),
    cond_var_summary = summary(object$cond.var),
    duration = as.numeric(object$duration, units = "secs")
  )
  
  # Assign a class to the summary object so that print.summary.cocoResid is dispatched.
  class(res) <- "summary.cocoResid"
  return(res)
}

#' @exportS3Method
print.summary.cocoResid <- function(x, ...) {
  cat("---- Residual Analysis Summary ----\n")
  cat("Model Type: ", x$model_type, "\n")
  cat("Model Order: ", x$model_order, "\n")
  cat("Time Series Length: ", x$ts_length, "\n")
  cat("Number of Fitted Values: ", x$n_fitted, "\n")
  cat("Covariates: ", x$covariates, "\n\n")
  
  cat("--- Residuals ---\n")
  print(x$residuals_summary)
  
  cat("\n--- Pearson Residuals ---\n")
  print(x$pearson_residuals_summary)
  
  cat("\n--- Conditional Variance ---\n")
  print(x$cond_var_summary)
  
  cat("\nCalculation Duration: ", x$duration, " seconds\n", sep = "")
  cat("\n")
  invisible(x)
}


#' @exportS3Method
fitted.cocoResid <- function(object, ...) {
  object$fitted
}

#' @exportS3Method
residuals.cocoResid <- function(object, ...) {
  object$residuals
}

dGP2 <- function(x, y, z, par) {
  if ((round(y) != y) | (y < 0)) {
    stop("y must be a non-negative integer")
  }

  if ((round(z) != z) | (z < 0)) {
    stop("z must be a non-negative integer")
  }

  if (round(x) != x) {
    stop("x must be an integer")
  }

  if (length(par) != 5) {
    stop("five parameters, lambda, alpha1, alpha2, alpha3 and eta, must be passed to the function")
  }

  lambda <- par[1]
  alpha1 <- par[2]
  alpha2 <- par[3]
  alpha3 <- par[4]
  eta <- par[5]

  if (lambda < 0) {
    stop("lambda must be greater than zero")
  }

  if ((alpha1 < 0) | (alpha1 > 1) | (eta < 0) | (eta > 1) | (alpha2 < 0) | (alpha2 > 1) | (alpha3 < 0) | (alpha3 > 1)) {
    stop("alpha and eta must be between or equal to zero and one")
  }



  if (x < 0) {
    result <- 0
  }

  if (x >= 0) {
    result <- dGP2h(x, y, z, lambda, alpha1, alpha2, alpha3, eta)
  }
  return(result)
}

#' @importFrom stats rpois
#' @importFrom stats runif
cocoSim_cov <- function(type, order, par, size, xreg, seasonality = c(1, 2), init = NULL, link_function="log") {
  
  if (length(seasonality == 1)) {
    seasonality <- c(seasonality, seasonality + 1)
  }
  
  if ((type != "GP") & (type != "Poisson")) {
    stop("Option 'type' must be either Poisson or GP")
  }

  if ((order != 1) & order != (2)) {
    stop("Option 'order' must be 1 or 2")
  }

  if (!is.numeric(par)) {
    stop("Option 'par' must be a numeric vector")
  }

  if (any(par[-c((length(par) - ncol(xreg)):length(par))] < 0) |
    any(par[-c((length(par) - ncol(xreg)):length(par))] > 1)) {
    stop("The autoregressive parameters and eta must be between zero and one")
  }

  if (length(seasonality) == 2) {
    if (seasonality[1] >= seasonality[2]) {
      stop("The first parameter of 'seasonality' must not be smaller than the second parameter")
    }

    if ((seasonality[2] != round(seasonality[2])) | (seasonality[2] < 1)) {
      stop("The values of 'seasonality' must be positive integer values")
    }
  }

  if ((seasonality[1] != round(seasonality[1])) | (seasonality[1] < 1)) {
    stop("The values of 'seasonality' must be positive integer values")
  }

  if (!is.numeric(size)) {
    stop("The value of 'size' must be a positive integer value")
  }


  if ((size != round(size)) | (size < 1)) {
    stop("The value of 'size' must be a positive integer value")
  }

  start_time <- Sys.time()

  T <- size


  # 1 Poisson 1
  if ((order == 1) & (type == "Poisson")) {
    if (length(par) != 1 + ncol(xreg)) {
      stop(paste("Number of parameters must equal", 1 + ncol(xreg), "for the Poisson 1 model with", ncol(xreg), "covariates"))
    }
    alpha <- par[1]
    eta <- 0


    vec_lambda <- c()
    for (j in 1:ncol(xreg)) {
      nam <- paste("lambda", j, sep = "")
      vec_lambda[j] <- assign(nam, par[j + 1])
    }

    data <- c()
    for (t in 1:seasonality[1]) {
      lambda_start_1 <- as.numeric(as.vector(xreg[t, ])) %*% vec_lambda
      data[t] <- stats::rpois(n = 1, applyLinkFunction(lambda_start_1, link_function))
    }
    
    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    } 
    
    lambda_raw <- xreg %*% vec_lambda
    lambdas <- apply(lambda_raw, 1, function(x) applyLinkFunction(x, link_function))
    
    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- c()
    for (index in 1:(length(lambdas))) {
      lambda <- lambdas[index]
      innovations[index] <- stats::rpois(n = 1, lambda)
    }
    uniform <- stats::runif(n = T, 0, 1)

    data <- simGP1cov(20, alpha, eta, vec_lambda, T, N, seasonality[1], data, xreg, uniform, innovations, link_function)

    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end Poisson 1



  # 2 GP1
  if ((order == 1) & (type == "GP")) {
    if (length(par) != 2 + ncol(xreg)) {
      stop(paste("Number of parameters must equal", 2 + ncol(xreg), "for the GP 1 model with", ncol(xreg), "covariates"))
    }
    alpha <- par[1]
    eta <- par[2]


    vec_lambda <- c()
    for (j in 1:ncol(xreg)) {
      nam <- paste("lambda", j, sep = "")
      vec_lambda[j] <- assign(nam, par[j + 2])
    }

    data <- c()
    for (t in 1:seasonality[1]) {
      lambda_start1 <- as.numeric(as.vector(xreg[t, ])) %*% vec_lambda
      data[t] <- rgenpois(n = 1, applyLinkFunction(lambda_start1, link_function),
                          eta)
    }

    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    } 

    lambda_raw <- xreg %*% vec_lambda
    lambdas <- apply(lambda_raw, 1, function(x) applyLinkFunction(x, link_function))

    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- c()
    for (index in 1:(length(lambdas))) {
      lambda <- lambdas[index]
      innovations[index] <- rgenpois(n = 1, lambda, eta)
    }
    uniform <- stats::runif(n = T, 0, 1)

    data <- simGP1cov(20, alpha, eta, vec_lambda, T, N, seasonality[1], data,
                      xreg, uniform, innovations, link_function)


    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end GP1


  # 3 Poisson 2
  if ((order == 2) & (type == "Poisson")) {
    if (length(par) != 3 + ncol(xreg)) {
      stop(paste("Number of parameters must equal", 3 + ncol(xreg), "for the Poisson 2 model with", ncol(xreg), "covariates"))
    }
    alpha1 <- par[1]
    alpha2 <- par[2]
    alpha3 <- par[3]
    eta <- 0
    U <- (1 - alpha1 - alpha2 - alpha3)^(-1)


    vec_lambda <- c()
    for (j in 1:ncol(xreg)) {
      nam <- paste("lambda", j, sep = "")
      vec_lambda[j] <- assign(nam, par[j + 3])
    }

    data <- c()
    for (t in 1:seasonality[2]) {
      if (nrow(xreg) > 1){
        lambda_start1 <- as.numeric(as.vector(xreg[t, ])) %*% vec_lambda
      } else{
        lambda_start1 <- c(0,0)
      }
      data[t] <- stats::rpois(n = 1, applyLinkFunction(lambda_start1, link_function))
    }

    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    } 

    lambda_raw <- xreg %*% vec_lambda
    lambdas <- apply(lambda_raw, 1, function(x) applyLinkFunction(x, link_function))

    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- c()
    for (index in 1:(length(lambdas))) {
      lambda <- lambdas[index]
      innovations[index] <- stats::rpois(n = 1, lambda)
    }
    if (nrow(xreg) < 2){
      uniform <- stats::runif(n = T+2, 0, 1)
    } else{
      uniform <- stats::runif(n = T, 0, 1)
    }

    data <- simGP2cov(20, alpha1, alpha2, alpha3, eta, vec_lambda, T, N, seasonality[1], seasonality[2], data, xreg, uniform, innovations, link_function)


    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end P2



  # 4 GP 2
  if ((order == 2) & (type == "GP")) {
    if (length(par) != 4 + ncol(xreg)) {
      stop(paste("Number of parameters must equal", 4 + ncol(xreg), "for the GP 2 model with", ncol(xreg), "covariates"))
    }

    alpha1 <- par[1]
    alpha2 <- par[2]
    alpha3 <- par[3]
    eta <- par[4]


    vec_lambda <- c()
    for (j in 1:ncol(xreg)) {
      nam <- paste("lambda", j, sep = "")
      vec_lambda[j] <- assign(nam, par[j + 4])
    }

    data <- c()
    for (t in 1:seasonality[2]) {
      if (nrow(xreg) > 1){
        lambda_start1 <- as.numeric(as.vector(xreg[t, ])) %*% vec_lambda
      } else{
        lambda_start1 <- c(0,0)
      }
      data[t] <- stats::rpois(n = 1, applyLinkFunction(lambda_start1, link_function))
    }

    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    } 

    lambda_raw <- xreg %*% vec_lambda
    lambdas <- apply(lambda_raw, 1, function(x) applyLinkFunction(x, link_function))

    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- c()
    for (index in 1:(length(lambdas))) {
      lambda <- lambdas[index]
      innovations[index] <- rgenpois(n = 1, lambda, eta)
    }
    
    if (nrow(xreg) < 3){
      uniform <- stats::runif(n = T+2, 0, 1)
    } else{
      uniform <- stats::runif(n = T, 0, 1)
    }
    
    data <- simGP2cov(20, alpha1, alpha2, alpha3, eta, vec_lambda, T, N,
                      seasonality[1], seasonality[2],
                      data, xreg, uniform, innovations, link_function)

    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end GP2

  end_time <- Sys.time()
  time <- end_time - start_time
  
  if (is.null(init)){
    warning("No burn-in period is specified using the init argument. Hence, the resulting simulated time series might not be stationary. You can add a custom burn-in period by passing it to the init argument. This could be, for example, done by simulating a burn-in period with appropriate covariates using cocoSim and passing the resulting time series to the init argument of a new cocoSim run.") 
  } else {
    output$data <- output$data[(order + 1):length(output$data)]
  }
  
  output$link_function <- link_function
  
  return(output)
}

#' @title Probability Integral Transform Based Model Assessment Procedure
#' @description Computes the probability integral transform (PIT) and provides
#' the non-randomized PIT histogram for assessing absolute performance of a
#' fitted model as proposed by Czado et al. (2009).
#' @param coco An object of class coco
#' @param J Number of bins for the histogram (default: 10)
#' @param conf.alpha Significance level for the confidence intervals (default: 0.05)
#' @param julia  if TRUE, the PIT is computed with \proglang{julia} (default: FALSE)
#' @return an object of class cocoPit. It contains the probability integral
#' transform values,  p-value of the chi-square goodness of fit test and information on the model specifications.
#' @details The adequacy of a distributional assumption for a model is assessed by
#' checking the cumulative non-randomized PIT distribution for uniformity.
#' A useful graphical device is the PIT histogram, which displays this
#' distribution to J equally spaced bins. We supplement the graph by
#' incorporating approximately \eqn{100(1 - \alpha)\%} confidence intervals obtained
#' from a standard chi-square goodness-of-fit test of the null hypothesis that
#' the J bins of the histogram are drawn from a uniform distribution.
#' For details, see Jung, McCabe and Tremayne (2016).
#' 
#' @importFrom stats pchisq
#' @importFrom stats qchisq
#' @importFrom ggplot2 autoplot
#' 
#' @references 
#' Czado, C., Gneiting, T. and Held, L. (2009) Predictive model assessment for count data. \emph{Biometrics} \bold{65}, 1254--61.
#' 
#' Jung, R. C., McCabe, B.P.M. and Tremayne, A.R. (2016). Model validation and diagnostics. \emph{In Handbook of Discrete
#' Valued Time Series}. Edited by Davis, R.A., Holan, S.H., Lund, R. and Ravishanker, N.. Boca Raton: Chapman and
#' Hall, pp. 189--218.
#' 
#' Jung, R. C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 3, 268--280.
#' @author Manuel Huth
#' @examples
#' lambda <- 1
#' alpha <- 0.4
#' set.seed(12345)
#' data <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)
#' fit <- cocoReg(order = 1, type = "Poisson", data = data)
#'
#' #PIT R implementation
#' pit_r <- cocoPit(fit)
#' plot(pit_r)
#' @export

cocoPit <- function(coco, J = 10, conf.alpha = 0.05, julia=FALSE) {
  start.time <- Sys.time()
  alpha <- conf.alpha

  if ((J != round(J)) | (J < 1)) {
    stop("The value of J must be a positive integer")
  }
  
  if (!is.null(coco$julia_reg) & julia){
    data <- coco$ts
    addJuliaFunctions()
    coco_pit <- JuliaConnectoR::juliaGet( JuliaConnectoR::juliaCall("cocoPit", coco$julia_reg, J))
    J_test <- J
    u <- coco_pit$values[[2]]
    d <- coco_pit$values[[1]]
  } else {

  data <- coco$ts
  seasonality <- c(1, 2) #will be used as argument in future versions coco$seasonality

  if ( is.null(coco$cov) ) {
    if ((coco$type == "GP") & (coco$order == 2)) {
      par <- coco$par
      lambda <- rep(par[1], length(data))
      alpha1 <- par[2]
      alpha2 <- par[3]
      alpha3 <- par[4]
      eta <- par[5]
      U <- 1 / (1 - alpha1 - alpha2 - alpha3)
      xreg <- NULL
    }

    if ((coco$type == "Poisson") & (coco$order == 2)) {
      par <- coco$par
      lambda <- rep(par[1], length(data))
      alpha1 <- par[2]
      alpha2 <- par[3]
      alpha3 <- par[4]
      U <- 1 / (1 - alpha1 - alpha2 - alpha3)
      xreg <- NULL
    }

    if ((coco$type == "GP") & (coco$order == 1)) {
      par <- coco$par
      lambda <- rep(par[1], length(data))
      alpha <- par[2]
      eta <- par[3]
      xreg <- NULL
    }

    if ((coco$type == "Poisson") & (coco$order == 1)) {
      par <- coco$par
      lambda <- rep(par[1], length(data))
      alpha <- par[2]
      xreg <- NULL
    }
  }



  if ( !is.null(coco$cov) ) {
    if ((coco$type == "Poisson") & (coco$order == 1)) {
      par <- coco$par
      alpha <- par[1]

      vec_lambda <- par[-(1)]
      xreg <- coco$cov


      # set up values for lambda
      lambda <- c()
      for (j in 1:length(data)) {
        lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
      }
    }

    if ((coco$type == "GP") & (coco$order == 1)) {
      par <- coco$par
      alpha <- par[1]
      eta <- par[2]

      vec_lambda <- par[-(1:2)]
      xreg <- coco$cov


      lambda <- c()
      for (j in 1:length(data)) {
        lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
      }
    }

    if ((coco$type == "Poisson") & (coco$order == 2)) {
      par <- coco$par
      alpha1 <- par[1]
      alpha2 <- par[2]
      alpha3 <- par[3]
      vec_lambda <- par[-(1:3)]
      xreg <- coco$cov

      U <- 1 / (1 - alpha1 - alpha2 - alpha3)

      lambda <- c()
      for (j in 1:length(data)) {
        lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
      }
    }

    if ((coco$type == "GP") & (coco$order == 2)) {
      par <- coco$par
      alpha1 <- par[1]
      alpha2 <- par[2]
      alpha3 <- par[3]
      eta <- par[4]

      vec_lambda <- par[-(1:4)]
      xreg <- coco$cov

      U <- 1 / (1 - alpha1 - alpha2 - alpha3)

      lambda <- c()
      for (j in 1:length(data)) {
        lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
      }
    }
  }





  ### PIT
  T <- length(data)
  J_test <- J
  J <- J + 1 # add 1 because the first bin is not used
  meanPIT <- matrix(0, nrow = T, ncol = J)

  for (t in (seasonality[2] + 1):T) {
    if ((coco$type == "Poisson") & (coco$order == 1)) {
      part <- c(lambda[t], alpha, 0)
      Px <- pGP1(data[t], data[t - seasonality[1]], part)
      if (data[t] >= 1) {
        Px1 <- pGP1(data[t] - 1, data[t - seasonality[1]], part)
      }
      if (data[t] < 1) {
        Px1 <- 0
      }
    }

    if ((coco$type == "GP") & (coco$order == 1)) {
      part <- c(lambda[t], alpha, eta)
      Px <- pGP1(data[t], data[t - seasonality[1]], part)
      if (data[t] >= 1) {
        Px1 <- pGP1(data[t] - 1, data[t - seasonality[1]], part)
      }
      if (data[t] < 1) {
        Px1 <- 0
      }
    }

    if ((coco$type == "Poisson") & (coco$order == 2)) {
      part <- c(lambda[t], alpha1, alpha2, alpha3, 0)
      Px <- pGP2(data[t], data[t - seasonality[1]], data[t - seasonality[2]], part)
      if (data[t] >= 1) {
        Px1 <- pGP2(data[t] - 1, data[t - seasonality[1]], data[t - seasonality[2]], part)
      }
      if (data[t] < 1) {
        Px1 <- 0
      }
    }

    if ((coco$type == "GP") & (coco$order == 2)) {
      part <- c(lambda[t], alpha1, alpha2, alpha3, eta)
      Px <- pGP2(data[t], data[t - seasonality[1]], data[t - seasonality[2]], part)
      if (data[t] >= 1) {
        Px1 <- pGP2(data[t] - 1, data[t - seasonality[1]], data[t - seasonality[2]], part)
      }
      if (data[t] < 1) {
        Px1 <- 0
      }
    }

    u <- seq(0, 1, length = J)
    N <- 1
    PIT <- matrix(0, nrow = N, ncol = J)
    for (i in 1:N) {
      temp <- numeric(J)
      for (s in 1:length(u)) {
        temp[s] <- (u[s] - Px1[i]) / (Px[i] - Px1[i])
        if (is.na(temp[s])) temp[s] <- 1
        if (temp[s] < 0) temp[s] <- 0
        if (temp[s] > 1) temp[s] <- 1
      }
      PIT[i, ] <- temp
    }
    meanPIT[t, ] <- colMeans(PIT)
  } # end t

  PIT <- colMeans(meanPIT)
  d <- diff(PIT)
  u <- u[-c(1)]
  } #end julia
  
  pval <- stats::pchisq(sum((d * length(data) - length(data) / J_test)^2 / (length(data) / J_test)), J_test-1)
  diff <- (sqrt(stats::qchisq(1-alpha, J_test-1)/J_test^2*length(data)) ) / length(data)
  confidence_band_values <- 1 / J_test + c(-1, 1) * diff 
  
  end.time <- Sys.time()

  time <- end.time - start.time
  list_out <- list(
    "type" = coco$type, "order" = coco$order, "ts" = coco$ts, "alpha" = alpha,
    "par" = par, "PIT values" = d, "duration" = time, "p_value" = pval,
    "confidence_bands" = confidence_band_values, "J" = J
  )
  
  class(list_out) <- "cocoPit"
  return(list_out)
}

#' @importFrom ggplot2 autoplot
#' @export
ggplot2::autoplot
#' @exportS3Method
autoplot.cocoPit <- function(object, ...){
  
  df <- data.frame(object$`PIT values`, 1:rep(length(object$`PIT values`)),
                   rep(object$confidence_bands[1], length(object$`PIT values`)),
                   rep(object$confidence_bands[2], length(object$`PIT values`)))
  colnames(df) <- c("pit", "bins", "lower", "upper")
  df_bands <- df
  df_bands$bins[1] <- df$bins[1] - 0.25
  df_bands$bins[length(df$bins)] <- df$bins[length(df$bins)] + 0.25
  pl <- ggplot2::ggplot(data=df, mapping = ggplot2::aes_string(x = "bins", y = "pit")) +
    ggplot2::geom_bar(stat="identity", position="dodge", width=0.3) + 
    ggplot2::labs(title = "Pit histogram", x = "Bins", y = "") +
    ggplot2::scale_x_continuous(breaks=df$bins) +
    ggplot2::theme_bw() + ggplot2::theme(text = ggplot2::element_text(size = 20)) +
    ggplot2::geom_ribbon(data=df_bands, ggplot2::aes_string(ymin="lower",ymax="upper"),
                         fill="steelblue", alpha=0.3) +
    ggplot2::ylim(c(0, min(1.05, max(df$pit)*1.5)))
  pl
}

#' @exportS3Method
plot.cocoPit <- function(x, ...) {
  p <- autoplot(
    x,
    ...
  )
  print(p)
}

#'@exportS3Method
print.cocoPit <- function(x, ...) {
  print(autoplot(x, ...,))
  invisible(x)
}

#' @exportS3Method
print.cocoPit <- function(x, ...) {
  cat("cocoPit Object")
  cat("\nModel Type: ", x$type, sep = "")
  cat("\nModel Order: ", x$order, sep = "")
  cat("\nPIT p-value: ", x$p_value, sep = "")
  cat("\nNumber of bins: ", x$J-1, sep = "")
  cat("\nBootstrapping Duration: ", as.numeric(x$duration, units = "secs"), " seconds", sep = "")
  cat("\n")
  
  invisible(x)
}

#' @exportS3Method
summary.cocoPit <- function(object, ...) {
  sum_obj <- list(
    model_type          = object$type,
    model_order         = object$order,
    ts_length           = length(object$ts),
    significance_level  = object$alpha,
    pit_values_summary  = summary(object[["PIT values"]]),
    confidence_bands    = object$confidence_bands,
    n_bins              = object$J - 1,
    pit_pvalue          = object$p_value
  )
  
  class(sum_obj) <- "summary.cocoPit"
  return(sum_obj)
}

#' @exportS3Method
print.summary.cocoPit <- function(x, ...) {
  cat("---- PIT Assessment Summary ----\n")
  cat("Model Type:          ", x$model_type, "\n")
  cat("Model Order:         ", x$model_order, "\n")
  cat("Time Series Length:  ", x$ts_length, "\n")
  cat("Significance Level:  ", x$significance_level, "\n\n")
  
  cat("Confidence Bands:    [", x$confidence_bands[1], ", ", x$confidence_bands[2], "]\n", sep = "")
  cat("Number of bins:      ", x$n_bins, "\n\n")
  
  cat("PIT Values Summary:\n")
  print(x$pit_values_summary)
  cat("\nPIT p-value:         ", x$pit_pvalue, "\n")
  
  invisible(x)
}

#' @importFrom forecast Acf
#' @importFrom stats glm lm constrOptim optim var na.omit
#' @importFrom JuliaConnectoR juliaCall
#' @importFrom numDeriv grad hessian
cocoReg_cov <- function(type, order, data, xreg, seasonality = c(1, 2), mu = 1e-4,
                        outer.it = 500, outer.eps = 1e-10, optim_control = FALSE, constrained.optim = TRUE, b.beta = -10,
                        start = NULL, start.val.adjust = TRUE, method_optim= "Nelder-Mead",
                        replace.start.val = 1e-5, iteration.start.val = 0.99,
                        method.hessian = "Richardson", julia_installed=FALSE, link_function="log", ...) {
  start_time <- Sys.time()

  if (is.data.frame(data)) {
    data <- as.matrix(data)
  }

  if (is.data.frame(xreg)) {
    data <- as.matrix(xreg)
  }

  if ((outer.it <= 0) | (outer.it != round(outer.it))) {
    stop("Option 'outer.it' must be a positive integer value")
  }

  if (outer.eps <= 0) {
    stop("Option 'outer.eps' must be a non-negative real number")
  }

  if (mu <= 0) {
    stop("Option 'mu' must be a non-negative real number")
  }

  if (replace.start.val <= 0) {
    stop("Option 'replace.start.val' must be a non-negative real number")
  }

  if ((type != "GP") & (type != "Poisson")) {
    stop("Option 'type' must be either Poisson or GP")
  }

  if ((constrained.optim != TRUE) & (constrained.optim != FALSE)) {
    stop("Option 'constrained.optim' must be either TRUE or FALSE")
  }

  if ((start.val.adjust != TRUE) & (start.val.adjust != FALSE)) {
    stop("Option 'start.val.adjust' must be either TRUE or False")
  }


  if ((order != 1) & order != (2)) {
    stop("Option 'order' must be 1 or 2")
  }
  
  if (length(seasonality == 1)) {
    seasonality <- c(seasonality, seasonality + 1)
  }

  if (seasonality[1] >= seasonality[2]) {
    stop("The first parameter of 'seasonality' must not be bigger than the second parameter")
  }

  if ((seasonality[1] != round(seasonality[1])) | (seasonality[1] < 1) |
    (seasonality[2] != round(seasonality[2])) | (seasonality[2] < 1)) {
    stop("The values of 'seasonality' must be positive integer values")
  }


  if (isTRUE(optim_control)) {
    optim_control <- list(trace = 1)
  } else {
    optim_control <- list()
  }
  se_boundary <- 0.00001
  ncov <- ncol(xreg)
  lag.max <- max(seasonality) + 1
  unconstrained.optim.lower <- -Inf
  unconstrained.optim.upper <- Inf
  
  df_covariates <- as.data.frame(cbind(data,xreg))
  
  if (link_function=="log"){
    covariates_glm_fit <- stats::glm(data ~ -1 + ., family="poisson", data=df_covariates)
    starting_values_covariates <- stats::na.omit(covariates_glm_fit$coefficients)
  } else if (link_function %in% c("identity", "relu", "softplus")) {
    covariates_glm_fit <- stats::lm(data ~ -1 + ., data=df_covariates)
    starting_values_covariates <- stats::na.omit(covariates_glm_fit$coefficients)
    starting_values_covariates[starting_values_covariates <= 0] <- 1e-5
  } else {
    stop("Choose either identity, log, relu, or softplus as link function.")
  }

  # PAR1
  if ((type == "Poisson") & (order == 1)) {
    PAR1 <- function(data) {
      mlf <- function(par, data) {
        alpha <- par[1]

        eta <- 0
        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 1])
        }

        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP1cov(20, alpha, eta, vec_lambda, T, seasonality[1], data, xreg_matrix, link_function)
        return(mlef)
      } # end function


      ## starting values
      if (length(start) == 0) {
        acf <- forecast::Acf(data, plot = FALSE, lag.max = lag.max)
        alpha_s <- acf$acf[seasonality[1] + 1]
        if (start.val.adjust == TRUE) {
          if (alpha_s < 0) {
            alpha_s <- replace.start.val
            warning(paste("The initial value of alpha is not in the feasible region and was set to", replace.start.val, ""))
          }
        }

        eta_s <- 0

        sta <- c(alpha_s, starting_values_covariates )

      }

      if (length(start) != 0) {
        sta <- start
        if (length(start) != 1 + ncol(xreg)) {
          stop(paste("Number of initial starting values must equal", 1 + ncol(xreg), "for the Poisson 1 model with", ncol(xreg), "covariates"))
        }
      }

      if (constrained.optim == TRUE) {
        # constrained.optimaints
        ui <- rbind(c(1), c(-1))

        matrix1 <- matrix(0, 2, ncol(xreg) ) # for additional columns

        matrix2 <- matrix(0, ncol(xreg) , 1) # for additional rows
        einheitsmatrix <- diag(ncol(xreg) )
        lower_part <- cbind(matrix2, einheitsmatrix)

        ui <- cbind(ui, matrix1) # right part of ne matrix
        ui <- rbind(ui, lower_part)

        ci <- c(0, -1, rep(b.beta, ncol(xreg)))

        # constrained.optimained optimization process
        if (method_optim == "BFGS") {
          gradient <- function(par, data){return(numDeriv::grad(func=mlf, x=par, data=data))}
          fit <- stats::constrOptim(
            theta = sta, f = mlf, grad=gradient, ui = ui, ci = ci, data = data, #mu = mu,
            method = method_optim, control = optim_control,
            #outer.iterations = outer.it, outer.eps = outer.eps,
            hessian = FALSE)          
        }else{
        fit <- stats::constrOptim(
          theta = sta, f = mlf, ui = ui, ci = ci, data = data, mu = mu,
          method = "Nelder-Mead", control = optim_control,
          outer.iterations = outer.it, outer.eps = outer.eps,
          hessian = FALSE
        )
        }
      } # end constrained.optim True

      if (constrained.optim == FALSE) {
        if (length(unconstrained.optim.lower) != 1 + ncol(xreg) & (unconstrained.optim.lower != -Inf)) {
          stop(paste("Number of lower bounds must equal", 1 + ncol(xreg), "for the Poisson 1 model with", ncol(xreg), "covariates"))
        }

        if (length(unconstrained.optim.upper) != 1 + ncol(xreg) & (unconstrained.optim.upper != Inf)) {
          stop(paste("Number of upper bounds must equal", 1 + ncol(xreg), "for the Poisson 1 model with", ncol(xreg), "covariates"))
        }

        fit <- stats::optim(
          par = sta, fn = mlf, gr = NULL, method = c("Nelder-Mead"), data = data,
          lower = unconstrained.optim.lower, upper = unconstrained.optim.upper,
          control = optim_control, hessian = FALSE
        )
      } # end constrained.optim False

      pars <- as.vector(unlist(fit$par, use.names = FALSE))
      lambs <- c()
      for (j in 1:ncov) {
        lambs[j] <- paste("beta", j, sep = "")
      }
      names(pars) <- c("alpha", lambs)
      T <- length(data)
      xreg_matrix <- data.matrix(xreg)
      likelihood <- -likelihoodGP1cov(20, pars[1], 0, pars[-c(1, 2)], T, seasonality[1], data, xreg_matrix, link_function)

      h <- function(par) {
        alpha <- par[1]

        eta <- 0

        data <- data
        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 1])
        }

        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP1cov(20, alpha, eta, vec_lambda, T, seasonality[1], data, xreg_matrix, link_function)
        return(mlef)
      } # end ml function


      gra <- numDeriv::grad(func = h, x = pars, method = method.hessian)
      hes <- numDeriv::hessian(func = h, x = pars, method = method.hessian)
      inv_hes <- solve(hes)
      se <- diag(inv_hes)^0.5
      names(se) <- c("alpha", lambs)

      end_time <- Sys.time()
      if ((pars[1] < se_boundary) |  (pars[1] > 1 - se_boundary)) {
        warning("The estimate of alpha is close to the boundary. Standard errors might not be valid.")
      }
      
      if (julia_installed) {
        addJuliaFunctions()
        julia_reg <- JuliaConnectoR::juliaCall("Coconots.create_julia_dict", 
                                               list("parameter", "covariance_matrix", "log_likelihood",
                                                    "type", "order", "data", "covariates",
                                                    "link", "starting_values", "optimizer", 
                                                    "lower_bounds", "upper_bounds", "optimization", 
                                                    "max_loop"),
                                               list(pars, inv_hes, likelihood,
                                                    type, order, data, xreg,
                                                    link_function, NULL, NULL, NULL, NULL,
                                                    NULL, NULL))
      } else {julia_reg = NULL}
      
      
      list_func <- list(
        "par" = pars, "gradient" = gra,
        "hessian" = hes, "inv hessian" = inv_hes, "se" = se,
        "ts" = data, "cov" = xreg, "type" = "Poisson", "order" = 1,
        "seasonality" = seasonality, "likelihood" = likelihood,
        "duration" = end_time - start_time, julia_reg = julia_reg, "link_function"=link_function
      )
      
      return(list_func)
    } # end PAR1
    return(PAR1(data))
  } # end if PAR1



  #GP1
  if ((type == "GP") & (order == 1)) {
    GP1 <- function(data) {
      mlf <- function(par, data) {
        alpha <- par[1]
        eta <- par[2]

        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 2])
        }

        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP1cov(20, alpha, eta, vec_lambda, T, seasonality[1], data, xreg_matrix, link_function)
        return(mlef)
      } # end function


      ## starting values
      if (length(start) == 0) {
        acf <- forecast::Acf(data, plot = FALSE, lag.max = lag.max)
        alpha_s <- acf$acf[seasonality[1] + 1]
        if (start.val.adjust == TRUE) {
          if (alpha_s < 0) {
            alpha_s <- replace.start.val
            warning(paste("The initial value of alpha is not in the feasible region and was set to", replace.start.val, ""))
          }
        }
        eta_s <- 1 - (mean(data) / stats::var(data))^0.5
        if (start.val.adjust == TRUE) {
          if (eta_s < 0) {
            eta_s <- replace.start.val
            warning(paste("The initial value of eta is not in the feasible region and was set to", replace.start.val, ""))
          }
          if (eta_s > 1) {
            eta_s <- 1 - replace.start.val
            warning(paste("The initial value of eta is not in the feasible region and was set to", 1 - replace.start.val, ""))
          }
        }

        sta <- c(alpha_s, eta_s, starting_values_covariates)

      }

      if (length(start) != 0) {
        sta <- start
        if (length(start) != 2 + ncol(xreg)) {
          stop(paste("Number of initial starting values must equal", 2 + ncol(xreg), "for the GP 1 model with", ncol(xreg), "covariates"))
        }
      }

      if (constrained.optim == TRUE) {
        # constrained.optimaints
        ui <- rbind(c(1, 0), c(-1, 0), c(0, 1), c(0, -1))

        matrix1 <- matrix(0, 4, ncol(xreg) ) # for additional columns

        matrix2 <- matrix(0, ncol(xreg) , 2) # for additional rows
        einheitsmatrix <- diag(ncol(xreg) )
        lower_part <- cbind(matrix2, einheitsmatrix)

        ui <- cbind(ui, matrix1) # right part of the matrix
        ui <- rbind(ui, lower_part)

        ci <- c(0, -1, 0, -1, rep(b.beta, ncol(xreg) ))

        # constrained.optimained optimization process
        if (method_optim == "BFGS") {
          gradient <- function(par, data){return(numDeriv::grad(func=mlf, x=par, data=data))}
          fit <- stats::constrOptim(
            theta = sta, f = mlf, grad=gradient, ui = ui, ci = ci, data = data, #mu = mu,
            method = method_optim, control = optim_control,
            #outer.iterations = outer.it, outer.eps = outer.eps,
            hessian = FALSE)          
        }else{
          fit <- stats::constrOptim(
            theta = sta, f = mlf, ui = ui, ci = ci, data = data, mu = mu,
            method = "Nelder-Mead", control = optim_control,
            outer.iterations = outer.it, outer.eps = outer.eps,
            hessian = FALSE
          )
        }
      } # constrained.optim True

      if (constrained.optim == FALSE) {
        if (length(unconstrained.optim.lower) != 2 + ncol(xreg) & (unconstrained.optim.lower != -Inf)) {
          stop(paste("Number of lower bounds must equal", 2 + ncol(xreg), "for the GP 1 model with", ncol(xreg), "covariates"))
        }

        if (length(unconstrained.optim.upper) != 2 + ncol(xreg) & (unconstrained.optim.upper != Inf)) {
          stop(paste("Number of upper bounds must equal", 2 + ncol(xreg), "for the GP 1 model with", ncol(xreg), "covariates"))
        }

        fit <- stats::optim(
          par = sta, fn = mlf, gr = NULL, method = c("Nelder-Mead"), data = data,
          lower = unconstrained.optim.lower, upper = unconstrained.optim.upper,
          control = optim_control, hessian = FALSE
        )
      } # end constrained.optim False

      pars <- as.vector(unlist(fit$par, use.names = FALSE))
      lambs <- c()
      for (j in 1:ncov) {
        lambs[j ] <- paste("beta", j, sep = "")
      }
      names(pars) <- c("alpha", "eta", lambs)
      T <- length(data)
      xreg_matrix <- data.matrix(xreg)
      likelihood <- -likelihoodGP1cov(20, pars[1], pars[2], pars[-c(1, 2)], T, seasonality[1], data, xreg_matrix, link_function)

      h <- function(par) {
        alpha <- par[1]
        eta <- par[2]

        data <- data
        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 2])
        }

        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP1cov(20, alpha, eta, vec_lambda, T, seasonality[1], data, xreg_matrix, link_function)
        return(mlef)
      } # end ml function




      gra <- numDeriv::grad(func = h, x = pars, method = method.hessian)
      hes <- numDeriv::hessian(func = h, x = pars, method = method.hessian)
      inv_hes <- solve(hes)
      se <- diag(inv_hes)^0.5
      names(se) <- c("alpha", "eta", lambs)

      end_time <- Sys.time()
      
      if ((pars[1] < se_boundary) |  (pars[1] > 1 - se_boundary)) {
        warning("The estimate of alpha is close to the boundary. Standard errors might not be valid.")
      }
      if ((pars[2] < se_boundary) |  (pars[2] > 1 - se_boundary)) {
        warning("The estimate of eta is close to the boundary. Standard errors might not be valid.")
      }
      
      if (julia_installed) {
        addJuliaFunctions()
        julia_reg <- JuliaConnectoR::juliaCall("Coconots.create_julia_dict", 
                                               list("parameter", "covariance_matrix", "log_likelihood",
                                                    "type", "order", "data", "covariates",
                                                    "link", "starting_values", "optimizer", 
                                                    "lower_bounds", "upper_bounds", "optimization", 
                                                    "max_loop"),
                                               list(pars, inv_hes, likelihood,
                                                    type, order, data, xreg,
                                                    link_function, NULL, NULL, NULL, NULL,
                                                    NULL, NULL))
      } else {julia_reg = NULL}
      
      list_func <- list(
        "par" = pars, "gradient" = gra,
        "hessian" = hes, "inv hessian" = inv_hes, "se" = se,
        "ts" = data, "cov" = xreg, "type" = "GP", "order" = 1,
        "seasonality" = seasonality, "likelihood" = likelihood, "duration" = end_time - start_time,
        julia_reg = julia_reg, "link_function"=link_function
      )
 
      return(list_func)
    } # end PGP1
    return(GP1(data))
  } # end if GP1


  #PAR2
  if ((type == "Poisson") & (order == 2)) {
    PAR2 <- function(data) {
      mlf <- function(par, data) {
        xreg <- xreg
        alpha1 <- par[1]
        alpha2 <- par[2]
        alpha3 <- par[3]

        eta <- 0

        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 3])
        }

        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP2cov(20, alpha1, alpha2, alpha3, eta, vec_lambda, T, seasonality[1], seasonality[2], data, xreg_matrix, link_function)

        return(mlef)
      } # end function

      ## starting values
      if (length(start) == 0) {
        acf <- forecast::Acf(data, plot = FALSE, lag.max = lag.max)
        p1 <- acf$acf[seasonality[1] + 1]
        p2 <- acf$acf[seasonality[2] + 1]

        v <- c()
        for (t in (seasonality[2] + 1):length(data)) {
          v[t - seasonality[2]] <- (data[t] - mean(data)) * (data[t - seasonality[1]] - mean(data)) * (data[t - seasonality[2]] - mean(data))
        }
        mu111 <- sum(v) / length(data)


        eta_s <- 0
        alpha3_s <- mu111 * (1 - eta_s)^4 / (mean(data) * (1 + 2 * eta_s))

        if (start.val.adjust == TRUE) {
          if (alpha3_s < 0) {
            alpha3_s <- replace.start.val
            warning(paste("The initial value of alpha3 is not in the feasible region and was set to", replace.start.val, ""))
          }
        }
        alpha2_s <- p2 - alpha3_s
        if (start.val.adjust == TRUE) {
          if (alpha2_s < 0) {
            alpha2_s <- replace.start.val
            warning(paste("The initial value of alpha2 is not in the feasible region and was set to", replace.start.val, ""))
          }
        }
        alpha1_s <- p1 - alpha3_s
        if (start.val.adjust == TRUE) {
          if (alpha1_s < 0) {
            alpha1_s <- replace.start.val
            warning(paste("The initial value of alpha1 is not in the feasible region and was set to", replace.start.val, ""))
          }

          if (alpha1_s + alpha2_s + alpha3_s > 1) {
            while (alpha1_s + alpha2_s + alpha3_s > 1) {
              alpha1_s <- alpha1_s * iteration.start.val
              alpha2_s <- alpha2_s * iteration.start.val
              alpha3_2 <- alpha3_s * iteration.start.val
            }
            warning(paste("The initial sum of 2*alpha1, alpha2 and alpha3 had not been in the feasible region and was adjusted", replace.start.val, ""))
          }


          if (2 * alpha1_s + alpha3_s > 1) {
            while (2 * alpha1_s + alpha3_s > 1) {
              alpha1_s <- alpha1_s * iteration.start.val
              alpha3_2 <- alpha3_s * iteration.start.val
            }
            warning(paste("The initial sum of alpha1 and alpha3 had not been in the feasible region and was adjusted", replace.start.val, ""))
          }
        }
        #starting values covariates

        sta <- c(alpha1_s, alpha2_s, alpha3_s, starting_values_covariates)
      }


      if (length(start) != 0) {
        sta <- start
        if (length(start) != 3 + ncol(xreg)) {
          stop(paste("Number of initial starting values must equal", 3 + ncol(xreg), "for the Poisson 2 model with", ncol(xreg), "covariates"))
        }
      }

      if (constrained.optim == TRUE) {
        # constrained.optimaints
        ui <- rbind(c(-1, -1, -1), c(-2, 0, -1), c(1, 0, 0), c(0, 1, 0), c(0, 0, 1))

        matrix1 <- matrix(0, 5, ncol(xreg) ) # for additional columns

        matrix2 <- matrix(0, ncol(xreg) , 3) # for additional rows
        einheitsmatrix <- diag(ncol(xreg) )
        lower_part <- cbind(matrix2, einheitsmatrix)

        ui <- cbind(ui, matrix1) # right part of ne matrix
        ui <- rbind(ui, lower_part)

        ci <- c(-1, -1, 0, 0, 0, rep(b.beta, ncol(xreg) ))

        # constrained.optimained optimization process
        if (method_optim == "BFGS") {
          gradient <- function(par, data){return(numDeriv::grad(func=mlf, x=par, data=data))}
          fit <- stats::constrOptim(
            theta = sta, f = mlf, grad=gradient, ui = ui, ci = ci, data = data, #mu = mu,
            method = method_optim, control = optim_control,
            #outer.iterations = outer.it, outer.eps = outer.eps,
            hessian = FALSE)          
        }else{
          fit <- stats::constrOptim(
            theta = sta, f = mlf, ui = ui, ci = ci, data = data, mu = mu,
            method = "Nelder-Mead", control = optim_control,
            outer.iterations = outer.it, outer.eps = outer.eps,
            hessian = FALSE
          )
        }
      } # end constrained.optim

      if (constrained.optim == FALSE) {
        if (length(unconstrained.optim.lower) != 3 + ncol(xreg) & (unconstrained.optim.lower != -Inf)) {
          stop(paste("Number of lower bounds must equal", 4 + ncol(xreg), "for the Poisson 2 model with", ncol(xreg), "covariates"))
        }

        if (length(unconstrained.optim.upper) != 3 + ncol(xreg) & (unconstrained.optim.upper != Inf)) {
          stop(paste("Number of upper bounds must equal", 3 + ncol(xreg), "for the Poisson 2 model with", ncol(xreg), "covariates"))
        }

        fit <- stats::optim(
          par = sta, fn = mlf, gr = NULL, method = c("Nelder-Mead"), data = data,
          lower = unconstrained.optim.lower, upper = unconstrained.optim.upper,
          control = optim_control, hessian = FALSE
        )
      } # end constrained.optim False

      pars <- as.vector(unlist(fit$par, use.names = FALSE))
      lambs <- c()
      for (j in 1:ncov) {
        lambs[j] <- paste("beta", j, sep = "")
      }
      names(pars) <- c("alpha1", "alpha2", "alpha3", lambs)
      T <- length(data)
      xreg_matrix <- data.matrix(xreg)
      likelihood <- -likelihoodGP2cov(20, pars[1], pars[2], pars[3], 0, pars[-c(1:4)], T, seasonality[1], seasonality[2], data, xreg_matrix, link_function)


      h <- function(par) {
        data <- data

        alpha1 <- par[1]
        alpha2 <- par[2]
        alpha3 <- par[3]

        eta <- 0

        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 3])
        }

        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP2cov(20, alpha1, alpha2, alpha3, eta, vec_lambda, T, seasonality[1], seasonality[2], data, xreg_matrix, link_function)
        return(mlef)
      } # end ml function

      gra <- numDeriv::grad(func = h, x = pars, method = method.hessian)
      hes <- numDeriv::hessian(func = h, x = pars, method = method.hessian)
      inv_hes <- solve(hes)
      se <- diag(inv_hes)^0.5
      names(se) <- c("alpha1", "alpha2", "alpha3", lambs)

      end_time <- Sys.time()
      
      if ((pars[1] < se_boundary) |  (pars[1] > 1 - se_boundary)) {
        warning("The estimate of alpha1 is close to the boundary. Standard errors might not be valid.")
      }
      if ((pars[2] < se_boundary) |  (pars[2] > 1 - se_boundary)) {
        warning("The estimate of alpha2 is close to the boundary. Standard errors might not be valid.")
      }
      if ((pars[3] < se_boundary) |  (pars[3] > 1 - se_boundary)) {
        warning("The estimate of alpha3 is close to the boundary. Standard errors might not be valid.")
      }
      
      if (julia_installed) {
        addJuliaFunctions()
        julia_reg <- JuliaConnectoR::juliaCall("Coconots.create_julia_dict", 
                                               list("parameter", "covariance_matrix", "log_likelihood",
                                                    "type", "order", "data", "covariates",
                                                    "link", "starting_values", "optimizer", 
                                                    "lower_bounds", "upper_bounds", "optimization", 
                                                    "max_loop"),
                                               list(pars, inv_hes, likelihood,
                                                    type, order, data, xreg,
                                                    link_function, NULL, NULL, NULL, NULL,
                                                    NULL, NULL))
      } else {julia_reg = NULL}
      
      list_func <- list(
        "par" = pars,
        "gradient" = gra, "hessian" = hes, "inv hessian" = inv_hes,
        "se" = se, "ts" = data, "cov" = xreg, "type" = "Poisson",
        "order" = 2, "seasonality" = seasonality, "likelihood" = likelihood,
        "duration" = end_time - start_time, julia_reg = julia_reg, "link_function"=link_function
      )

      return(list_func)
    } # end PAR2
    return(PAR2(data))
  } # end if PAR2


  # GP2
  if ((type == "GP") & (order == 2)) {
    GP2 <- function(data) {
      mlf <- function(par, data) {
        xreg <- xreg
        alpha1 <- par[1]
        alpha2 <- par[2]
        alpha3 <- par[3]
        eta <- par[4]
        U <- (1 - alpha1 - alpha2 - alpha3)^(-1)

        ### 1
        #### define lambda_t
        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 4])
        }
        vec_lambda <- vec_lambda


        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP2cov(20, alpha1, alpha2, alpha3, eta, vec_lambda, T, seasonality[1], seasonality[2], data, xreg_matrix, link_function)

        return(mlef)
      } # end function

      ## starting values
      if (length(start) == 0) {
        acf <- forecast::Acf(data, plot = FALSE, lag.max = lag.max)
        p1 <- acf$acf[[seasonality[1] + 1]]
        p2 <- acf$acf[[seasonality[2] + 1]]

        v <- c()
        for (t in (seasonality[2] + 1):length(data)) {
          v[t - seasonality[2]] <- (data[t] - mean(data)) * (data[t - seasonality[1]] - mean(data)) * (data[t - seasonality[2]] - mean(data))
        }
        mu111 <- sum(v) / length(data)


        eta_s <- 1 - (mean(data) / stats::var(data))^0.5
        if (start.val.adjust == TRUE) {
          if (eta_s < 0) {
            eta_s <- replace.start.val
            warning(paste("The initial value of eta is not in the feasible region and was set to", replace.start.val, ""))
          }

          if (eta_s > 1) {
            eta_s <- 1 - replace.start.val
            warning(paste("The initial value of eta is not in the feasible region and was set to", 1 - replace.start.val, ""))
          }
        }

        alpha3_s <- mu111 * (1 - eta_s)^4 / (mean(data) * (1 + 2 * eta_s))
        if (start.val.adjust == TRUE) {
          if (alpha3_s < 0) {
            alpha3_s <- replace.start.val
            warning(paste("The initial value of alpha3 is not in the feasible region and was set to", replace.start.val, ""))
          }
        }

        alpha2_s <- p2 - alpha3_s
        if (start.val.adjust == TRUE) {
          if (alpha2_s < 0) {
            alpha2_s <- replace.start.val
            warning(paste("The initial value of alpha2 is not in the feasible region and was set to", replace.start.val, ""))
          }

          if (alpha2_s > 1) {
            alpha2_s <- 1 - replace.start.val
            warning(paste("The initial value of alpha2 is not in the feasible region and was set to", 1 - replace.start.val, ""))
          }
        }

        alpha1_s <- p1 - alpha3_s
        if (start.val.adjust == TRUE) {
          if (alpha1_s < 0) {
            alpha1_s <- replace.start.val
            warning(paste("The initial value of alpha1 is not in the feasible region and was set to", replace.start.val, ""))
          }

          if (alpha1_s + alpha2_s + alpha3_s > 1) {
            while (alpha1_s + alpha2_s + alpha3_s > 1) {
              alpha1_s <- alpha1_s * iteration.start.val
              alpha2_s <- alpha2_s * iteration.start.val
              alpha3_2 <- alpha3_s * iteration.start.val
            }
            warning(paste("The initial sum of 2*alpha1, alpha2 and alpha3 had not been in the feasible region and was adjusted", ""))
          }


          if (2 * alpha1_s + alpha3_s > 1) {
            while (2 * alpha1_s + alpha3_s > 1) {
              alpha1_s <- alpha1_s * iteration.start.val
              alpha3_2 <- alpha3_s * iteration.start.val
            }
            warning(paste("The initial sum of alpha1 and alpha3 had not been in the feasible region and was adjusted", ""))
          }
        }

        lambda_s <- mean(data) * (1 - eta_s) * (1 - alpha1_s - alpha2_s - alpha3_s)
        if (start.val.adjust == TRUE) {
          if (lambda_s < 0) {
            lambda_s <- replace.start.val
            warning(paste("The initial value of lambda is not in the feasible region and was set to", replace.start.val, ""))
          }
        }

        sta <- c(alpha1_s, alpha2_s, alpha3_s, eta_s, starting_values_covariates)

      }

      if (length(start) != 0) {
        sta <- start
        if (length(start) != 4 + ncol(xreg)) {
          stop(paste("Number of initial starting values must equal", 4 + ncol(xreg), "for the GP 2 model with", ncol(xreg), "covariates"))
        }
      }


      if (constrained.optim == TRUE) {
        # constrained.optimaints
        ui <- rbind(
          c(-1, -1, -1, 0), c(-2, 0, -1, 0), c(1, 0, 0, 0), c(0, 1, 0, 0), c(0, 0, 1, 0),
          c(0, 0, 0, 1), c(0, 0, 0, -1)
        )
        matrix1 <- matrix(0, 7, ncol(xreg)) # for additional columns

        matrix2 <- matrix(0, ncol(xreg) , 4) # for additional rows
        einheitsmatrix <- diag(ncol(xreg) )
        lower_part <- cbind(matrix2, einheitsmatrix)

        ui <- cbind(ui, matrix1) # right part of the matrix
        ui <- rbind(ui, lower_part)

        ci <- c(-1, -1, 0, 0, 0, 0, -1, rep(b.beta, ncol(xreg) ))

        # constrained.optimained optimization process
        if (method_optim == "BFGS") {
          gradient <- function(par, data){return(numDeriv::grad(func=mlf, x=par, data=data))}
          fit <- stats::constrOptim(
            theta = sta, f = mlf, grad=gradient, ui = ui, ci = ci, data = data, #mu = mu,
            method = method_optim, control = optim_control,
            #outer.iterations = outer.it, outer.eps = outer.eps,
            hessian = FALSE)          
        }else{
          fit <- stats::constrOptim(
            theta = sta, f = mlf, ui = ui, ci = ci, data = data, #mu = mu,
            method = "Nelder-Mead", control = optim_control,
            outer.iterations = outer.it, outer.eps = outer.eps,
            hessian = FALSE
          )
        }
      } # end constrained.optim True

      if (constrained.optim == FALSE) {
        if (length(unconstrained.optim.lower) != 4 + ncol(xreg) & (unconstrained.optim.lower != -Inf)) {
          stop(paste("Number of lower bounds must equal", 4 + ncol(xreg), "for the GP 2 model with", ncol(xreg), "covariates"))
        }

        if (length(unconstrained.optim.upper) != 4 + ncol(xreg) & (unconstrained.optim.upper != Inf)) {
          stop(paste("Number of upper bounds must equal", 4 + ncol(xreg), "for the GP 2 model with", ncol(xreg), "covariates"))
        }

        fit <- stats::optim(
          par = sta, fn = mlf, gr = NULL, method = c("Nelder-Mead"), data = data,
          lower = unconstrained.optim.lower, upper = unconstrained.optim.upper,
          control = optim_control, hessian = FALSE
        )
      } # end constrained.optim False

      pars <- as.vector(unlist(fit$par, use.names = FALSE))
      lambs <- c()
      for (j in 1:ncov) {
        lambs[j] <- paste("beta", j, sep = "")
      }
      names(pars) <- c("alpha1", "alpha2", "alpha3", "eta", lambs)
      T <- length(data)
      xreg_matrix <- data.matrix(xreg)
      likelihood <- -likelihoodGP2cov(20, pars[1], pars[2], pars[3], pars[4], pars[-c(1:4)], T, seasonality[1], seasonality[2], data, xreg_matrix, link_function)


      h <- function(par) {
        data <- data


        alpha1 <- par[1]
        alpha2 <- par[2]
        alpha3 <- par[3]
        eta <- par[4]
        U <- (1 - alpha1 - alpha2 - alpha3)^(-1)
        ### define lambda_t
        vec_lambda <- c()
        for (j in 1:ncol(xreg)) {
          nam <- paste("lambda", j, sep = "")
          vec_lambda[j] <- assign(nam, par[j + 4])
        }

        T <- length(data)
        xreg_matrix <- data.matrix(xreg)
        mlef <- likelihoodGP2cov(20, alpha1, alpha2, alpha3, eta, vec_lambda, T, seasonality[1], seasonality[2], data, xreg_matrix, link_function)


        return(mlef)
      } # end ml function

      gra <- numDeriv::grad(func = h, x = pars, method = method.hessian)
      hes <- numDeriv::hessian(func = h, x = pars, method = method.hessian)
      inv_hes <- solve(hes)
      se <- diag(inv_hes)^0.5
      names(se) <- c("alpha1", "alpha2", "alpha3", "eta", lambs)

      end_time <- Sys.time()
      
      if ((pars[1] < se_boundary) |  (pars[1] > 1 - se_boundary)) {
        warning("The estimate of alpha1 is close to the boundary. Standard errors might not be valid.")
      }
      if ((pars[2] < se_boundary) |  (pars[2] > 1 - se_boundary)) {
        warning("The estimate of alpha2 is close to the boundary. Standard errors might not be valid.")
      }
      if ((pars[3] < se_boundary) |  (pars[3] > 1 - se_boundary)) {
        warning("The estimate of alpha3 is close to the boundary. Standard errors might not be valid.")
      }
      
      if ((pars[4] < se_boundary) |  (pars[4] > 1 - se_boundary)) {
        warning("The estimate of eta is close to the boundary. Standard errors might not be valid.")
      }
      
      if (julia_installed) {
        addJuliaFunctions()
        julia_reg <- JuliaConnectoR::juliaCall("Coconots.create_julia_dict", 
                                               list("parameter", "covariance_matrix", "log_likelihood",
                                                    "type", "order", "data", "covariates",
                                                    "link", "starting_values", "optimizer", 
                                                    "lower_bounds", "upper_bounds", "optimization", 
                                                    "max_loop"),
                                               list(pars, inv_hes, likelihood,
                                                    type, order, data, xreg,
                                                    link_function, NULL, NULL, NULL, NULL,
                                                    NULL, NULL))
      } else {julia_reg = NULL}
      
      list_func <- list(
        "par" = pars,
        "gradient" = gra, "hessian" = hes, "inv hessian" = inv_hes,
        "se" = se, "ts" = data, "cov" = xreg, "type" = "GP", "order" = 2,
        "seasonality" = seasonality, "likelihood" = likelihood,
        "duration" = end_time - start_time, julia_reg = julia_reg, "link_function"=link_function
      )

      return(list_func)
    } # end GP2
    return(GP2(data))
  } # end if GP2
}

#' @importFrom JuliaConnectoR juliaCall
cocoRegJulia <- function(type, order, data, xreg, start, link_function, lower_bound_covariates){
  addJuliaFunctions()
  return(JuliaConnectoR::juliaCall("cocoReg", type, order, data, xreg, start, link_function, lower_bound_covariates))
}

#' @title Computes Scores for Various Models Maintaining a Common Sample 
#'
#' @description This function computes log, quadrtic and ranked probability scores for Poisson and Generalized Poisson models.
#'
#' @param data A numeric vector containing the data to be used for modeling
#' @param models A character string specifying which models to use. Default is `"all"`, which uses both Poisson and GP models.
#' @param print.progress A logical value indicating whether to print progress messages (Default: `TRUE`).
#' @param max_x_score An integer which is used as the maximum count for the computation
#'  of the score (defaul: `50`)
#' @param julia if TRUE, \code{cocoSoc} is run with \proglang{julia} (default: FALSE)
#' @param ... Additional arguments to be passed to the `cocoReg` function.
#' 
#' @author Manuel Huth
#' 
#' @details Supports model selection by computing score over a range of models while maintaining a common sample and 
#' a common specification.
#' @return A list of class `"cocoSoc"` containing:
#' \describe{
#'   \item{fits}{A list of fitted model objects.}
#'   \item{scores_list}{A list of score objects for each model.}
#'   \item{scores_df}{A data frame containing the logarithmic, quadratic, and ranked probability scores for each model.}
#' }
#'
#'
#' @export
cocoSoc <- function(data, models = "all", print.progress=TRUE,
                       max_x_score=50, julia = FALSE, ...
          ) {
  fits <- list()
  scores <- list()
  
  n_models <- ifelse(models == "all", 2, 4)
  
  if (models == "all"){
    models_use <- c("Poisson", "GP")
  } else {
    models_use <- c(models)
  } 
  
  #exclude first observation for order 1 models
  index <- 1
  for (type in models_use){
    for (order in 1:2){
      data_use <- data[(order*-1 + 3):length(data)]
      if (type == "Poisson"){name <- "PAR"} else {name <- "GP"}
      
      if (print.progress){
        print(paste0("Computing ", name, order, " model."))
      }
      
      name_mod <- paste0(name, order)
      
      fits[[name_mod]] <- cocoReg(type, order, data_use, ...)

      scores[[name_mod]] <- cocoScore(fits[[name_mod]], max_x = max_x_score, julia=T)
      index <- index + 1
    }
  }
  
  log_score <- c()
  quad_score <- c()
  rps_score <- c()
  for (model in names(scores)){
    log_score <- c(log_score, scores[[model]]$log.score)
    quad_score <- c(quad_score, scores[[model]]$quad.score)
    rps_score <- c(rps_score, scores[[model]]$rps.score)
  }
  
  df <- data.frame(names(scores))
  colnames(df) <- c(" ")
  df[c("Logarithmic Score", "Quadratic Score", "Ranked Probability Score")] <- cbind(log_score, quad_score, rps_score)

  list_out <- list("fits" = fits, "scores_list" = scores, "scores_df" = df)
  class(list_out) <- "cocoSoc"
  
  return(list_out)

  } # end function

#' @exportS3Method
summary.cocoSoc <- function(object, ...) {
  class(object) <- "summary.cocoSoc"
  return(object)
}

#' @exportS3Method
print.summary.cocoSoc <- function(x, ...) {
  
  print(x$scores_df, print.gap=3, quote=FALSE, na.print="")
  
}

cocoForecastKSteps <- function(fit, k=3, number_simulations=1000, alpha=0.05, covariates=NULL,
                               decimals=4, 
                               julia=FALSE){
  if (julia){

    forecasts <- cocoForecastKStepsJulia(fit, k=k, number_simulations = number_simulations,
                                   covariates=covariates)
    
  } else {
    forecasts <- cocoForecastKStepsRCPP(fit, k=k, number_simulations = number_simulations,
                                  covariates=covariates)
  }
  
  make_class <- function(i){
    df_temp <- forecasts[[i]]
    df_temp$value <- as.numeric(df_temp$value)
    
    # Generate the full sequence of values
    all_values <- data.frame(value = 0:max(df_temp$value))
    
    # Merge with the original data, filling missing frequencies with 0
    df_complete <- merge(all_values, df_temp, by = "value", all.x = TRUE)
    df_complete$frequency[is.na(df_complete$frequency)] <- 0
    
    
    
    densities <- df_complete[,"frequency"]
    x <- df_complete[,"value"]
    
    
    mode <- match(max(densities), densities) - 1
    distribution_function <- cumsum(densities)
    median <- min(which(distribution_function >= 0.5)) - 1
    densities_plot <- round(densities, decimals)
    
    out <- list("density" = densities, "mode" = mode, mean=sum(densities*x),
                "median" = median, "densities_plot" = densities_plot, "x" = x,
                "k" = i,
                "lower"=min(which(distribution_function >= alpha/2)) - 1,
                "upper"=min(which(distribution_function >= 1-alpha/2)) - 1,
                "data"=fit$ts)
    class(out) <- "cocoForecast"
    return(out)
  }
  
  list_out <- lapply(1:length(forecasts), make_class)
  class(list_out) <- "cocoForecastCollection"
  return(list_out)
}

#' @importFrom JuliaConnectoR juliaCall
cocoForecastKStepsJulia <- function(fit, k=3, number_simulations=500, covariates=NULL){
  dict_out <- JuliaConnectoR::juliaCall("cocoPredictKsteps", fit$julia_reg, k, number_simulations, covariates, fit$link_function)
  return_frequencies <- function(i){as.data.frame(dict_out[[paste0("prediction_", i)]])}
  return(lapply(1:dict_out[["length"]], return_frequencies))
}

#' @title Bootstrap Based Model Assessment Procedure
#' @description Model checking procedure emphasising reproducibility in fitted models, as proposed by Tsay (1992).
#' @param coco An object of class coco
#' @param numb.lags Number of lags for which to compute sample autocorrelations (default: 21).
#' @param rep.Bootstrap Number of bootstrap replicates to use (default: 1000)
#' @param conf.alpha \eqn{100(1-\code{conf.alpha})\%} probability interval for the acceptance envelopes (default: 0.05)
#' @param julia  if TRUE, the bootstrap is run with \proglang{julia} (default: FALSE)
#' @param julia_seed Seed for the \proglang{julia} implementation. Only used if \proglang{julia} equals TRUE
#' @return an object of class cocoBoot. It contains the bootstrapped confidence intervals
#' of the autocorrelations and information on the model specifications.
#' @details Bootstrap-generated acceptance envelopes for the autocorrelation function provides an overall evaluation by comparing it with the sample autocorrelation function in a joint plot. 
#' @importFrom forecast Acf
#' @importFrom matrixStats rowQuantiles
#' @importFrom ggplot2 autoplot
#' @references 
#' Tsay, R. S. (1992) Model checking via parametric bootstraps in time series analysis. \emph{Applied Statistics} \bold{41}, 1--15.
#' @examples
#' lambda <- 1
#' alpha <- 0.4
#' set.seed(12345)
#' data <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)
#' fit <- cocoReg(order = 1, type = "Poisson", data = data)
#'
#' # bootstrap model assessment - R implementation
#' boot_r <- cocoBoot(fit, rep.Bootstrap=400)
#' plot(boot_r)
#' @export

cocoBoot <- function(coco, numb.lags = 21, rep.Bootstrap = 1000,
                 conf.alpha = 0.05, julia = FALSE, julia_seed = NULL
                 ) {
  start.time <- Sys.time()
  confidence <- 1 - conf.alpha
  if ((confidence <= 0) | (confidence >= 1)) {
    stop("Option confidence must be a real number between 0 and 1")
  }

  if ((numb.lags != round(numb.lags)) | (numb.lags < 1)) {
    stop("The value of numb.lags must be a positive integer")
  }

  if ((rep.Bootstrap != round(rep.Bootstrap)) | (rep.Bootstrap < 1)) {
    stop("The value of rep.Bootstrap must be a positive integer")
  }
  
  data <- coco$ts
  
  if (!is.null(coco$julia_reg) & julia){
    if (!is.null(julia_seed)){
      setJuliaSeed(julia_seed)
    }
    addJuliaFunctions()
    coco_boot <- JuliaConnectoR::juliaGet( JuliaConnectoR::juliaCall("cocoBoot", coco$julia_reg, 1:numb.lags, rep.Bootstrap, 1-confidence))
    #acfdata <- coco_boot$values[[2]]
    ac <- t(coco_boot$values[[1]]) 
    #confidence_bands <- data.frame(cbind(coco_boot$values[[3]], coco_boot$values[[4]]))
    #colnames(confidence_bands) <- c("upper", "lower")
  } else {
    
    seasonality <- c(1,2) #will be used as argument in future versions
  
    conf.alpha <- 1 - confidence
    
    if ( is.null(coco$cov) ) {
      if ((coco$type == "GP") & (coco$order == 2)) {
        par <- coco$par
        lambda <- par[1]
        alpha1 <- par[2]
        alpha2 <- par[3]
        alpha3 <- par[4]
        eta <- par[5]
        U <- 1 / (1 - alpha1 - alpha2 - alpha3)
        xreg <- NULL
      }
  
      if ((coco$type == "Poisson") & (coco$order == 2)) {
        par <- coco$par
        lambda <- par[1]
        alpha1 <- par[2]
        alpha2 <- par[3]
        alpha3 <- par[4]
        U <- 1 / (1 - alpha1 - alpha2 - alpha3)
        xreg <- NULL
      }
  
      if ((coco$type == "GP") & (coco$order == 1)) {
        par <- coco$par
        lambda <- par[1]
        alpha <- par[2]
        eta <- par[3]
        xreg <- NULL
      }
  
      if ((coco$type == "Poisson") & (coco$order == 1)) {
        par <- coco$par
        lambda <- par[1]
        alpha <- par[2]
        xreg <- NULL
      }
    }
  
  
    if ( !is.null(coco$cov) )  {
      if ((coco$type == "Poisson") & (coco$order == 1)) {
        par <- coco$par
        alpha <- par[1]
  
        vec_lambda <- par[-(1)]
        xreg <- coco$cov
        data <- coco$ts
  
        # set up values for lambda
        lambda <- c()
        for (j in 1:length(data)) {
          lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
        }
      }
  
      if ((coco$type == "GP") & (coco$order == 1)) {
        par <- coco$par
        alpha <- par[1]
        eta <- par[2]
        vec_lambda <- par[-(1:2)]
        xreg <- coco$cov
        data <- coco$ts
  
        lambda <- c()
        for (j in 1:length(data)) {
          lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
        }
      }
  
      if ((coco$type == "Poisson") & (coco$order == 2)) {
        par <- coco$par
        alpha1 <- par[1]
        alpha2 <- par[2]
        alpha3 <- par[3]
        vec_lambda <- par[-(1:3)]
        xreg <- coco$cov
        data <- coco$ts
        U <- 1 / (1 - alpha1 - alpha2 - alpha3)
  
        lambda <- c()
        for (j in 1:length(data)) {
          lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
        }
      }
  
      if ((coco$type == "GP") & (coco$order == 2)) {
        par <- coco$par
        alpha1 <- par[1]
        alpha2 <- par[2]
        alpha3 <- par[3]
        eta <- par[4]
  
        vec_lambda <- par[-(1:4)]
        xreg <- coco$cov
        data <- coco$ts
        U <- 1 / (1 - alpha1 - alpha2 - alpha3)
  
        lambda <- c()
        for (j in 1:length(data)) {
          lambda[j] <- applyLinkFunction(as.numeric(as.vector(xreg[j, ])) %*% vec_lambda, coco$link_function)
        }
      }
    }
  
  
    # Parametric Bootstrap
    T <- length(data)
    nlags <- numb.lags
    nB <- rep.Bootstrap
    B <- matrix(NaN, nrow = T, ncol = nB)
    ac <- matrix(NaN, nrow = nlags, ncol = nB)
    conf <- conf.alpha
  
    if ((is.null(coco$cov)) ) {
      for (b in 1:nB) {
        help <- cocoSim(order = coco$order, type = coco$type, par = par, length = (T + 10))[11:(T + 10)]
        B[, b] <- help
        ac[, b] <- forecast::Acf(help, plot = FALSE, lag.max = nlags)$acf[2:(nlags + 1)]
      }
    }
  
  
    if (!is.null(coco$cov)) {
      xreg <- as.matrix(xreg)
      for (b in 1:nB) {
        help <- cocoSim(type = coco$type, order = coco$order, par = par, length = T, xreg = xreg, link_function=coco$link_function)
        B[, b] <- help
        ac[, b] <- forecast::Acf(help, plot = FALSE, lag.max = nlags)$acf[2:(nlags + 1)]
      }
    }
  

  } #end julia
  acfdata <- forecast::Acf(data, plot = FALSE, lag.max = numb.lags)$acf[2:(numb.lags + 1)]
  confidence_bands <- data.frame(matrixStats::rowQuantiles(ac, probs = c((1-confidence)/2, 1-(1-confidence)/2)))
  df_plot <- cbind(acfdata, 1:numb.lags, confidence_bands)
  colnames(df_plot) <- c("y", "x", "lower", "upper")
  

  end.time <- Sys.time()
  time <- end.time - start.time
  list_out <- list(
    "type" = coco$type, "order" = coco$order, "ts" = coco$ts, "cov" = coco$cov, means = acfdata,
    "confidence" = confidence_bands, "duration" = time, "df_plot" = df_plot, "rep.Bootstrap" = rep.Bootstrap
  )
  class(list_out) <- "cocoBoot"
  return(list_out)
} # end function

#' @exportS3Method
print.cocoBoot <- function(x, ...) {
  cat("cocoBoot Object")
  cat("\nModel Type: ", x$type, sep = "")
  cat("\nModel Order: ", x$order, sep = "")
  cat("\nBootstrap Draws: ", as.numeric(x$rep.Bootstrap), sep = "")
  cat("\nBootstrapping Duration: ", as.numeric(x$duration, units = "secs"), " seconds\n", sep = "")
  cat("\n")
  
  invisible(x)
}

#' @importFrom ggplot2 autoplot
#' @export
ggplot2::autoplot
#' @exportS3Method
autoplot.cocoBoot <- function(object, ...){
  
  pl <- ggplot2::ggplot(object$df_plot, ggplot2::aes_string("x", "y")) +
    ggplot2::geom_point()+
    ggplot2::geom_line()+
    ggplot2::geom_ribbon(data=object$df_plot,ggplot2::aes_string(ymin="lower",ymax="upper"),
                         fill="steelblue", alpha=0.3) +
    ggplot2::theme_bw() + ggplot2::xlab("Lags") + ggplot2::ylab("Autocorrelation") +
    ggplot2::ggtitle("Parametric Bootstrap") +
    ggplot2::theme(text = ggplot2::element_text(size = 20))
  pl
  
}

#' @exportS3Method
plot.cocoBoot <- function(x, ...) {
  p <- autoplot(
    x,
    ...
  )
  print(p)
}

#' @exportS3Method
summary.cocoBoot <- function(object, ...) {
  sum_obj <- list(
    model_type    = object$type,
    model_order   = object$order,
    ts_length     = length(object$ts),
    covariates    = if (is.null(object$cov)) {
      "None"
    } else {
      if (is.data.frame(object$cov) || is.matrix(object$cov)) {
        paste("Present (", ncol(as.matrix(object$cov)), " variable(s))", sep = "")
      } else {
        "Present (1 variable)"
      }
    },
    out_of_region = sum(object$df_plot$y < object$df_plot$lower | object$df_plot$y > object$df_plot$upper),
    total_rows    = nrow(object$df_plot),
    bootstrap_draws = as.numeric(object$rep.Bootstrap),
    boot_duration = as.numeric(object$duration, units = "secs")
  )
  
  class(sum_obj) <- "summary.cocoBoot"
  return(sum_obj)
}

#' @exportS3Method
print.summary.cocoBoot <- function(x, ...) {
  cat("---- Bootstrap Assessment Summary ----\n")
  cat("Model Type:         ", x$model_type, "\n")
  cat("Model Order:        ", x$model_order, "\n")
  cat("Time Series Length: ", x$ts_length, "\n")
  cat("Covariates:         ", x$covariates, "\n")
  cat("Bootstrap Draws:    ", x$bootstrap_draws, "\n")
  cat("Bootstrapping Duration: ", x$boot_duration, " seconds\n", sep = "")
  cat("ACF values outside bootstrapped confidence region: ", 
      x$out_of_region, " out of ", x$total_rows, "\n")
  cat("\n")
  invisible(x)
}

#' @title K-Step Ahead Forecast Distributions
#' @description Computes the k-step ahead forecast (distributions) using the models in the coconots package. 
#' 
#' @param object An object that has been fitted previously, of class coco.
#' @param k The number of steps ahead for which the forecast should be computed (Default: 1).
#' @param number_simulations The number of simulation runs to compute (Default: 1000).
#' @param alpha Significance level used to construct the prediction intervals (Default: 0.05).
#' @param simulate_one_step_ahead If FALSE, the one-step ahead prediction is obtained using the analytical predictive distribution. If TRUE, bootstrapping is used.
#' @param max The maximum number of the forecast support for the plot. If NULL all values for which the cumulative distribution function is below 1- epsilon are used for the plot.
#' @param epsilon If max is NULL, epsilon determines the range of the support that is used by subsequent automatic plotting using R's plot() function.
#' @param xcast An optional matrix of covariate values for the forecasting. If `NULL`, the function assumes no covariates.
#' @param decimals Number of decimal places for the forecast probabilities
#' @param julia  if TRUE, the estimate is predicted with \proglang{julia} (Default: FALSE).
#' @param ... Optional arguments.
#'
#' @return A \code{cocoForecastCollection} object (list of \code{cocoForecast} objects, one per step ahead). Use \code{summary()} to obtain a data frame of point forecasts (mean, median, mode) and prediction intervals. Individual steps are accessible via \code{forecast[[i]]}.
#' 
#' @importFrom ggplot2 autoplot
#' @importFrom stats predict
#' @details Returns forecasts for each mass point of the k-step ahead
#' distribution for the fitted model. The exact predictive distributions for
#' one-step ahead predictions for
#' the models included here are provided in Jung and Tremayne (2011), maximum
#' likelihood estimates replace the true model parameters. 
#' For k>1 forecast distributions are estimated using a parametric bootstrap. See Jung and Tremanye (2006).
#' Out-of-sample values
#' for covariates can be provided, if necessary.
#' 
#' 
#' for k > 1 
#' 
#' @examples
#' length <- 500
#' pars <- c(1, 0.4)
#' set.seed(12345)
#' data <- cocoSim(order = 1, type = "Poisson", par = pars, length = length)
#' fit <- cocoReg(order = 1, type = "Poisson", data = data)
#' forecast <- predict(fit, k=1, simulate_one_step_ahead = FALSE)
#' plot(forecast[[1]]) #plot one-step ahead forecast distribution
#' 
#' @references 
#' Jung, R.C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 3, 268--280.
#' 
#' Jung, R.C. and Tremayne, A.R. (2006) Coherent forecasting in integer time series models. 
#'  \emph{International Journal of Forecasting} \bold{22}, 223--238
#' @rdname predict.coco
#' @exportS3Method
predict.coco <- function(object, k=1, number_simulations=1000,
                         alpha=0.05,
                         simulate_one_step_ahead = FALSE,
                         max=NULL, epsilon=1e-8, xcast=NULL,
                         decimals = 4,
                         julia=FALSE,...) {
  
  if ((k == 1) & (!is.matrix(xcast)) & (!is.null(xcast))){
    xcast <- matrix(xcast, nrow=1)
  }
  
  if ((k==1) & (!simulate_one_step_ahead)){
    return(cocoForecastOneStep(object, max=max, epsilon=epsilon, xcast=xcast,
                               alpha=alpha,
                    decimals=decimals, julia=julia))
  } else if ((k > 1) & (!simulate_one_step_ahead))  {
    
    k_steps <- cocoForecastKSteps(object, k=k, number_simulations=number_simulations,
                       covariates=xcast, alpha=alpha,
                       decimals=decimals, 
                       julia=julia)
    k_steps[[1]] <- cocoForecastOneStep(object, max=max, epsilon=epsilon, xcast=xcast,
                                        alpha=alpha,
                                        decimals=decimals, julia=julia)[[1]]
    return(k_steps)
  } else {
    return(cocoForecastKSteps(object, k=k, number_simulations=number_simulations,
                              covariates=xcast, alpha=alpha,
                              decimals=decimals, 
                              julia=julia))
  }
}


#' @importFrom ggplot2 autoplot
#' @export
ggplot2::autoplot


#' @exportS3Method
autoplot.cocoForecast <- function(object, width = 0.9, ...) {
  df <- data.frame(x = factor(object$x, levels = sort(unique(object$x))),
                   p = as.numeric(object$densities_plot))
  
  ggplot2::ggplot(df, ggplot2::aes(x = x, y = p)) +
    ggplot2::geom_col(width = width) +
    ggplot2::labs(title = "Probability Forecast", x = "Support", y = "Probability") +
    ggplot2::theme_bw() +
    ggplot2::theme(text = ggplot2::element_text(size = 20))
}


#' @exportS3Method
plot.cocoForecast <- function(x, ...) {
  p <- autoplot(
    x,
    ...
  )
  suppressWarnings({print(p)})
}

#' @importFrom ggplot2 autoplot
#' @export
ggplot2::autoplot
#' @exportS3Method
autoplot.cocoForecastCollection <- function(object, forecast_type="mode",
                                            n_lags_plot=30, ...){
  return_forecast <- function(i){
    L <- object[[i]]
    c(df_data[nrow(df_data), "t"]+i, NA, L[[forecast_type]], L[["lower"]], L[["upper"]])
  }
  
  relevant_data <- object[[1]]$data[max(1,(length(object[[1]]$data)-n_lags_plot+1)):length(object[[1]]$data)]
  df_data <- cbind(max(1,(length(object[[1]]$data)-n_lags_plot+1)):length(object[[1]]$data), relevant_data, NA, NA, NA)
  colnames(df_data) <- c("t", "x", "forecast", "lower", "upper")
  
  df <- as.data.frame(t(sapply(1:length(object), return_forecast)))
  colnames(df) <- c("t", "x", "forecast", "lower", "upper")
  
  df_plot <- rbind(df_data, df)
  
  df_transition <- df_plot[(nrow(df_plot)-length(object)):nrow(df_plot) ,]
  df_transition[1, "forecast"] <- df_transition[1, "x"] 
  
  pl <- ggplot2::ggplot(mapping = ggplot2::aes(x = df_plot[, "t"], y = df_plot[, "x"])) +
    ggplot2::geom_point(ggplot2::aes(color="Data")) + ggplot2::geom_line() + 
    ggplot2::geom_point(mapping = ggplot2::aes(x = df[, "t"], y = df[, "forecast"],
                                               color = "Prediction")) +
    ggplot2::geom_line(mapping = ggplot2::aes(x = df_transition[, "t"],
                                              y = df_transition[, "forecast"]),
                       color = "steelblue", linetype = "dotted") +
    ggplot2::geom_errorbar(mapping = ggplot2::aes(
      ymin = df_plot[, "lower"],
      ymax = df_plot[, "upper"]), color = "steelblue") +
    ggplot2::labs(title = paste0(forecast_type, " forecast"), x = "Time", y = "Count (x)") +
    ggplot2::theme_bw() + ggplot2::theme(text = ggplot2::element_text(size = 20)) +
    ggplot2::scale_color_manual(name='',
                                breaks=c('Data', 'Prediction'),
                                values=c('Data'='black', 'Prediction'='steelblue'))
  
  pl
}

#' @exportS3Method
plot.cocoForecastCollection <- function(x, ...) {
  p <- autoplot(
    x,
    ...
  )
  suppressWarnings({print(p)})
}

#'@exportS3Method
print.cocoForecastCollection <- function(x, ...) {
  suppressWarnings({print(autoplot(x, ...,))})
  invisible(x)
}

#' @exportS3Method
summary.cocoForecast <- function(object, ...) {
  out <- data.frame(
    step   = if (!is.null(object$k)) object$k else 1L,
    mean   = object$mean,
    median = object$median,
    mode   = object$mode,
    lower  = object$lower,
    upper  = object$upper
  )
  class(out) <- c("summary.cocoForecast", "data.frame")
  out
}

#' @exportS3Method
print.summary.cocoForecast <- function(x, ...) {
  cat("---- Forecast Summary ----\n")
  print.data.frame(x, row.names = FALSE, ...)
  invisible(x)
}

#' @exportS3Method
summary.cocoForecastCollection <- function(object, ...) {
  rows <- lapply(seq_along(object), function(i) {
    fc <- object[[i]]
    data.frame(
      step   = if (!is.null(fc$k)) fc$k else i,
      mean   = fc$mean,
      median = fc$median,
      mode   = fc$mode,
      lower  = fc$lower,
      upper  = fc$upper
    )
  })
  out <- do.call(rbind, rows)
  class(out) <- c("summary.cocoForecastCollection", "data.frame")
  out
}

#' @exportS3Method
print.summary.cocoForecastCollection <- function(x, ...) {
  cat("---- Forecast Summary ----\n")
  print.data.frame(x, row.names = FALSE, ...)
  invisible(x)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

fact <- function(i) {
    .Call(`_coconots_fact`, i)
}

applyLinkFunction <- function(mathelp, link_function) {
    .Call(`_coconots_applyLinkFunction`, mathelp, link_function)
}

innerProduct <- function(x, y) {
    .Call(`_coconots_innerProduct`, x, y)
}

likelihoodGP2 <- function(sumlimit, lambda, alpha1, alpha2, alpha3, eta, T, seas1, seas2, data) {
    .Call(`_coconots_likelihoodGP2`, sumlimit, lambda, alpha1, alpha2, alpha3, eta, T, seas1, seas2, data)
}

likelihoodGP1 <- function(sumlimit, lambda, alpha, eta, T, seas, data) {
    .Call(`_coconots_likelihoodGP1`, sumlimit, lambda, alpha, eta, T, seas, data)
}

simGP2 <- function(sumlimit, lambda, alpha1, alpha2, alpha3, eta, T, N, seas1, seas2, data, uniform, innovations) {
    .Call(`_coconots_simGP2`, sumlimit, lambda, alpha1, alpha2, alpha3, eta, T, N, seas1, seas2, data, uniform, innovations)
}

simGP1 <- function(sumlimit, lambda, alpha, eta, T, N, seas, data, uniform, innovations) {
    .Call(`_coconots_simGP1`, sumlimit, lambda, alpha, eta, T, N, seas, data, uniform, innovations)
}

dGP2h <- function(x, y, z, lambda, alpha1, alpha2, alpha3, eta) {
    .Call(`_coconots_dGP2h`, x, y, z, lambda, alpha1, alpha2, alpha3, eta)
}

dGP1h <- function(x, y, lambda, alpha, eta) {
    .Call(`_coconots_dGP1h`, x, y, lambda, alpha, eta)
}

dR2 <- function(r, y, z, lambda, alpha1, alpha2, alpha3, eta) {
    .Call(`_coconots_dR2`, r, y, z, lambda, alpha1, alpha2, alpha3, eta)
}

Pyz <- function(y, z, lambda, alpha1, alpha2, alpha3, eta) {
    .Call(`_coconots_Pyz`, y, z, lambda, alpha1, alpha2, alpha3, eta)
}

simGP2cov <- function(sumlimit, alpha1, alpha2, alpha3, eta, lambdas, T, N, seas1, seas2, data, xreg, uniform, innovations, link_function) {
    .Call(`_coconots_simGP2cov`, sumlimit, alpha1, alpha2, alpha3, eta, lambdas, T, N, seas1, seas2, data, xreg, uniform, innovations, link_function)
}

simGP1cov <- function(sumlimit, alpha, eta, lambdas, T, N, seas, data, xreg, uniform, innovations, link_function) {
    .Call(`_coconots_simGP1cov`, sumlimit, alpha, eta, lambdas, T, N, seas, data, xreg, uniform, innovations, link_function)
}

likelihoodGP2cov <- function(sumlimit, alpha1, alpha2, alpha3, eta, lambdas, T, seas1, seas2, data, xreg, link_function) {
    .Call(`_coconots_likelihoodGP2cov`, sumlimit, alpha1, alpha2, alpha3, eta, lambdas, T, seas1, seas2, data, xreg, link_function)
}

likelihoodGP1cov <- function(sumlimit, alpha, eta, lambdas, T, seas, data, xreg, link_function) {
    .Call(`_coconots_likelihoodGP1cov`, sumlimit, alpha, eta, lambdas, T, seas, data, xreg, link_function)
}

#' Time Series of Daily Downloads of a TeX-Editor
#'
#' The data represent the number of daily downloads of a TeX-editor between 
#' June 2006 and February 2007. The dataset contains 267 observations. 
#' The data have been provided by Christian Weiss.
#'
#' @source Weiss, C.H. (2008) 
#'   Thinning operations for modelling time series of counts -- a survey. 
#'   \emph{Advances in Statistical Analysis}, \bold{92}, 319--341.
#'
#' @format A time series (`ts`) object containing daily data from June 2006 to February 2007.
#' @keywords datasets
#' 
"downloads"

#' @title Set Seed for \proglang{julia}'s Random Number Generator
#' @description Sets the seed for \proglang{julia}'s random number generator to ensure reproducibility.
#' @param julia_seed An integer seed value to be passed to \proglang{julia}'s random number generator.
#' @details This function initializes the necessary \proglang{julia} functions and sets the random seed for \proglang{julia}. 
#' If the provided seed is NULL, the function does nothing.
#' @importFrom JuliaConnectoR juliaEval
#' @author Manuel Huth
#' @export
setJuliaSeed <- function(julia_seed){
  addJuliaFunctions()
  if (!is.null(julia_seed)){
    JuliaConnectoR::juliaEval(paste0('Random.seed!(Int64(', julia_seed, '))', collapse=""))
  }
}

dGP1 <- function(x, y, par) {
  if ((round(y) != y) | (y < 0)) {
    stop("y must be a non-negative integer")
  }

  if (round(x) != x) {
    stop("x must be an integer")
  }

  if (length(par) != 3) {
    stop("three parameters, lambda, alpha and eta, must be passed to the function")
  }

  lambda <- par[1]
  alpha <- par[2]
  eta <- par[3]

  if (lambda < 0) {
    stop("lambda must be greater than zero")
  }

  if ((alpha < 0) | (alpha > 1) | (eta < 0) | (eta > 1)) {
    stop("alpha and eta must be between or equal to zero and one")
  }


  if (x < 0) {
    result <- 0
  }

  if (x >= 0) {
    result <- dGP1h(x, y, lambda, alpha, eta)
  }

  return(result)
}

#' @importFrom JuliaConnectoR juliaGet
#' @importFrom utils head
cocoForecastOneStep <- function(coco, max=NULL, epsilon=1e-12, xcast=NULL,
                                alpha=0.05,
                         decimals = 4, julia=FALSE) {
  
  seasonality <- c(1, 2) #will be used as argument in future versions
  
  
  if (!is.vector(xcast) & !is.null(xcast)){
    if (nrow(xcast) > 1){
      xcast <- xcast[1, ]
    }
  }
  
  if (is.null(max)){
    max_use <- 60
  } else{
    max_use <- max
  }
  
  if (!is.null(coco$julia_reg) & julia){
    addJuliaFunctions()
    if ((is.matrix(xcast))) {
      xcast <- c(xcast)
    }
    coco_forecast <- JuliaConnectoR::juliaGet( JuliaConnectoR::juliaCall("cocoPredictOneStep",
                                                                         coco$julia_reg, 0:max_use, xcast))
    densities <- coco_forecast$values[[4]]
    if (is.null(max)){
      cumulative <- cumsum(densities)
      index_use <- min(which(cumulative >= 1-epsilon))
      densities <- densities[1:index_use]
    }
    
    mode <- coco_forecast$values[[1]]
    median <- coco_forecast$values[[2]]
  } else {
    if (length(seasonality == 1)) {
      seasonality = c(seasonality, seasonality + 1)
    }
    
    data <- coco$ts
    y <- data[length(data) - seasonality[1] + 1]
    z <- data[length(data) - seasonality[2] + 1]
    parameter <-coco$par
    
    if ( !is.null(xcast) ){ 
      number_covariates <- ncol(coco$cov) 
      betas <- parameter[(length(parameter)-number_covariates+1):length(parameter)]
      parameter <- utils::head(parameter, -number_covariates)
      dot_product <- betas %*% c(xcast)
      lambda <- applyLinkFunction(dot_product, coco$link_function)
      parameter <- c(lambda, parameter)
    }
    
    type <- coco$type
    order <- coco$order
    
    densities <- c()
    
    x <- 0:max_use
    
    if (order == 1){
      if (type == "Poisson"){parameter <- c(parameter,0)}
      for (number in x) { 
        densities[number+1] <- dGP1(x=number, y=y, par=parameter)
      }
    } 
    
    if (order == 2){
      if (type == "Poisson"){parameter <- c(parameter,0)}
      for (number in x) { 
        densities[number+1] <- dGP2(x=number, y=y, z=z, par=parameter)
      }
    }
    
    mode <- match(max(densities), densities) - 1
    distribution_function <- cumsum(densities)
    median <- min(which(distribution_function >= 0.5)) - 1
    if (is.null(max)){
      cumulative <- cumsum(densities)
      index_use <- min(which(cumulative >= 1-epsilon))
      densities <- densities[1:index_use]
    }
  }
  
  
  distribution_function <- cumsum(densities)
  densities_plot <- round(densities, decimals)
  
  x <- 0:(length(densities)-1)
  
  out <- list("density" = densities, "mode" = mode, mean=sum(densities*as.numeric(x)),
              "median" = median, "densities_plot" = densities_plot, "x" = x,
              "lower"=min(which(distribution_function >= alpha/2)) - 1,
              "upper"=min(which(distribution_function >= 1-alpha/2)) - 1,
              "data"=coco$ts)
  class(out) <- "cocoForecast"
  
  list_out <- list(out)
  class(list_out) <- "cocoForecastCollection"
  return(list_out) 
}

#' Time Series of Gold Particle Counts in a Well-Defined Colloidal Solution
#'
#' A sample of 370 counts of gold particles in a well-defined colloidal
#' solution at equidistant points in time. The data were originally 
#' published in Westgren (1916) and later used in Jung and Tremayne (2006).
#'
#' @source Jung, R.C.; Tremayne, A.R. (2006) 
#'   Coherent forecasting in integer time series models. 
#'   \emph{International Journal of Forecasting}, \bold{22}, 223--238.
#'
#'   Westgren, A. (1916) Die Veraenderungsgeschwindigkeit der lokalen 
#'   Teilchenkonzentration in kolloidalen Systemen (Erste Mitteilung). 
#'   \emph{Arkiv foer Matematik, Astronomi och Fysik}, \bold{11}, 1--24.
#'
#' @format A time series (`ts`) object containing 370 observations at equidistant time points.
#' @keywords datasets
#' @examples
#' plot(goldparticle)
"goldparticle"

#' @title Scoring Rule Based Model Assessment Procedure
#' @description The function computes log, quadratic and ranked probability scores for assessing relative performance of a fitted model.
#' @param coco An object of class coco
#' @param max_x An integer which is used as the maximum count for the computation
#'  of the score (default: `50`)
#' @param julia if TRUE, the scores are computed with \proglang{julia} (default: FALSE).
#' @return a list containing the log score, quadratic score and ranked probability score.
#' @details Scoring rules assign a numerical score based on the predictive
#'  distribution and the observed data  to measure the quality of probabilistic predictions.
#' They are provided here as a model selection tool and are computed as
#'  averages over the relevant set of (in-sample) predictions. Scoring rules are, generally, negatively oriented
#' penalties that one seeks to minimize. The literature has developed a large number of scoring
#' rules and, unless there is a unique and clearly defined underlying decision problem,
#' there is no automatic choice of a (proper) scoring rule to be used in any given situation.
#' Therefore, the use of a variety of scoring rules may be appropriate to take advantage of
#' specific emphases and strengths. Three proper scoring rules
#' (for a definition of the concept of propriety see Gneiting and Raftery, 2007),
#' which Jung, McCabe and Tremayne (2016) found to be particularly useful, are implemented.
#' For more information see the references listed below.
#' 
#' @importFrom JuliaConnectoR juliaGet
#' @references 
#' Czado, C. and Gneitling, T. and Held, L. (2009) Predictive Model Assessment for Count Data. \emph{Biometrics}, \bold{65},  1254--1261.
#'
#' Gneiting, T. and Raftery, A. E. (2007) Strictly proper scoring rules, prediction, and estimation. \emph{Journal
#' of the American Statistical Association}, 102:359-378.
#'
#' Jung, R. C., McCabe, B.P.M. and Tremayne, A.R. (2016). Model validation and diagnostics. \emph{In Handbook of Discrete
#' Valued Time Series}. Edited by Davis, R.A., Holan, S.H., Lund, R. and Ravishanker, N.. Boca Raton: Chapman and
#' Hall, pp. 189--218.
#' 
#' Jung, R. C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 268--280.
#' @author Manuel Huth
#' @examples
#' lambda <- 1
#' alpha <- 0.4
#' set.seed(12345)
#' data <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)
#' fit <- cocoReg(order = 1, type = "Poisson", data = data)
#'
#' # scoring rules - R implementation
#' score_r <- cocoScore(fit)
#' @export

cocoScore <- function(coco, max_x = 50, julia=FALSE) {
  
  
  if (!is.null(coco$julia_reg) & julia){
    addJuliaFunctions()
    coco_score <- JuliaConnectoR::juliaGet( JuliaConnectoR::juliaCall("compute_scores", coco$julia_reg, max_x))
    return(list("log.score" = coco_score$values[[1]], "quad.score" = coco_score$values[[2]],
                 "rps.score" = coco_score$values[[3]],
                "aic" = 2*length(coco$par) - 2 * coco$likelihood,
                "bic" = length(coco$par) * log(length(coco$ts)) - 2 * coco$likelihood
                ))
  } else {
  T <- length(coco$ts)
  par <- coco$par
  seas <- c(1, 2) #will be used as argument in future versions
  data <- coco$ts

  if (coco$type == "Poisson"){
    if (is.null(coco$cov)){
      par <- c(par, 0)
    } else {
      par <- c( par[1:(length(par)-ncol(coco$cov))], 0, par[(length(par)-ncol(coco$cov)+1):length(par)])
    }
  }
  
  if (!is.null(coco$cov)){
    xreg <- coco$cov
    betas <- par[(length(par)-ncol(xreg)+1):length(par)]
    lambda_raw <- xreg %*% betas
    lambdas <- apply(lambda_raw, 1, function(x) applyLinkFunction(x, coco$link_function))
    
    par_original <- par
    
  }
  
  #Poisson/GP 1
  if  (coco$order == 1){
    calculate_scores <- function(t) {
      if (!is.null(coco$cov)){
        par <- c(lambdas[t], par_original[1:(length(par_original) - ncol(xreg))]) 
      }
      
      # log score
      log_score <- -log(dGP1(data[t], data[t - seas[1]], par))
      
      # quadratic score
      norm_p <- sum(sapply(0:max_x, function(j) dGP1(j, data[t - seas[1]], par)^2))
      quad_score <- -2 * dGP1(data[t], data[t - seas[1]], par) + norm_p
      
      # ranked probability score
      sum_sq_diff <- sum(sapply(0:max_x, function(j) (pGP1(j, data[t - seas[1]], par) - (j >= data[t]))^2))
      rps_score <- sum_sq_diff
      
      return(c(log_score, quad_score, rps_score))
    }
    
    # Use sapply for the outer loop
    scores <- sapply((seas[1] + 1):T, calculate_scores)
    
    # Summarize scores
    log.score <- mean(scores[1, ])
    quad.score <- mean(scores[2, ])
    rps.score <- mean(scores[3, ])

  }# Poisson/GP 1


  #Poisson/GP 2
  if (coco$order == 2){
    calculate_scores <- function(t) {
      if (!is.null(coco$cov)){
        par <- c(lambdas[t], par_original[1:(length(par_original) - ncol(xreg))]) 
      }
      
      # log score
      log_score <- -log(dGP2(data[t], data[t - seas[1]], data[t - seas[2]], par))
      
      # quadratic score
      norm_p <- sum(sapply(0:max_x, function(j) dGP2(j, data[t - seas[1]], data[t - seas[2]], par)^2))
      quad_score <- -2 * dGP2(data[t], data[t - seas[1]], data[t - seas[2]], par) + norm_p
      
      # ranked probability score
      sum_sq_diff <- sum(sapply(0:max_x, function(j) (pGP2(j, data[t - seas[1]], data[t - seas[2]], par) - (j >= data[t]))^2))
      rps_score <- sum_sq_diff
      
      return(c(log_score, quad_score, rps_score))
    }
    
    # Use sapply for the outer loop
    scores <- sapply((seas[2] + 1):T, calculate_scores)
    
    # Summarize scores
    log.score <- mean(scores[1, ])
    quad.score <- mean(scores[2, ])
    rps.score <- mean(scores[3, ])
  }# Poisson/GP 2

  list_out <- list("log.score" = log.score, "quad.score" = quad.score, "rps.score" = rps.score)

  }#end julia if
  
  list_out["bic"] <- length(coco$par) * log(length(coco$ts)) - 2 * coco$likelihood
  list_out["aic"] <- 2*length(coco$par) - 2 * coco$likelihood
  return(list_out)

}# end function

#' @importFrom stats rpois
#' @importFrom stats runif
cocoSim_base <- function(type, order, par, size, seasonality = c(1, 2), init = NULL) {
  
  if (length(seasonality == 1)) {
    seasonality <- c(seasonality, seasonality + 1)
  }
  
  if ((type != "GP") & (type != "Poisson")) {
    stop("Option 'type' must be either Poisson or GP")
  }

  if ((order != 1) & order != (2)) {
    stop("Option 'order' must be 1 or 2")
  }

  if (!is.numeric(par)) {
    stop("Option 'par' must be a numeric vector")
  }

  if (par[1] <= 0) {
    stop("lambda must be a positive reel number")
  }

  if (!((all(par[-1] > 0)) & all(par[-1] < 1))) {
    stop(" The autoregressive parameters and eta must be between zero and one")
  }

  if (length(seasonality) == 2) {
    if (seasonality[1] >= seasonality[2]) {
      stop("The first parameter of 'seasonality' must not be smaller than the second parameter")
    }

    if ((seasonality[2] != round(seasonality[2])) | (seasonality[2] < 1)) {
      stop("The values of 'seasonality' must be positive integer values")
    }
  }

  if ((seasonality[1] != round(seasonality[1])) | (seasonality[1] < 1)) {
    stop("The values of 'seasonality' must be positive integer values")
  }

  if (!is.numeric(size)) {
    stop("The value of 'size' must be a positive integer value")
  }

  if ((size != round(size)) | (size < 1)) {
    stop("The value of 'size' must be a positive integer value")
  }

 



  start_time <- Sys.time()

  T <- size

  # 1 Poisson 1
  if ((order == 1) & (type == "Poisson")) {
    if (length(par) != 2) {
      stop(paste("Number of parameteres must equal 2 for the Poisson 1 model"))
    }

    lambda <- par[1]
    alpha <- par[2]
    eta <- 0
    psi <- eta * (1 - alpha) / lambda

    
    data <- stats::rpois(n = seasonality[1], lambda)
    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    } 
    
    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- stats::rpois(n = T, lambda)
    uniform <- stats::runif(n = T, 0, 1)

    data <- simGP1(20, lambda, alpha, eta, T, N, seasonality[1], data, uniform, innovations)


    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end Poisson 1



  # 2 GP1
  if ((order == 1) & (type == "GP")) {
    if (length(par) != 3) {
      stop(paste("Number of parameteres must equal 3 for the GP 1 model"))
    }
    lambda <- par[1]
    alpha <- par[2]
    eta <- par[3]

    data <- rgenpois(n = seasonality[1], lambda, eta)

    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    }

    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- rgenpois(n = T, lambda, eta)
    uniform <- stats::runif(n = T, 0, 1)

    data <- simGP1(20, lambda, alpha, eta, T, N, seasonality[1], data, uniform, innovations)

    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end GP1


  # 3 Poisson 2
  if ((order == 2) & (type == "Poisson")) {
    if (length(par) != 4) {
      stop(paste("Number of parameteres must equal 4 for the Poisson 2 model"))
    }
    lambda <- par[1]
    alpha1 <- par[2]
    alpha2 <- par[3]
    alpha3 <- par[4]
    eta <- 0

    data <- stats::rpois(n = seasonality[2], lambda)

    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    }

    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- stats::rpois(n = T, lambda)
    uniform <- stats::runif(n = T, 0, 1)
    data <- simGP2(
      20, lambda, alpha1, alpha2, alpha3, eta, T, N, seasonality[1],
      seasonality[2], data, uniform, innovations
    )

    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end P2



  # 4 GP2
  if ((order == 2) & (type == "GP")) {
    if (length(par) != 5) {
      stop(paste("Number of parameteres must equal 5 for the GP 2 model"))
    }
    lambda <- par[1]
    alpha1 <- par[2]
    alpha2 <- par[3]
    alpha3 <- par[4]
    eta <- par[5]

    data <- rgenpois(n = seasonality[2], lambda, eta)

    if (!is.null(init) ) {
      data <- init[(length(init) - length(data)+1):(length(init))]
    }

    N <- length(data)
    data <- c(data, rep(NaN, T - N))
    innovations <- rgenpois(n = T, lambda, eta)
    uniform <- stats::runif(n = T, 0, 1)
    data <- simGP2(
      20, lambda, alpha1, alpha2, alpha3, eta, T, N, seasonality[1],
      seasonality[2], data, uniform, innovations
    )

    end_time <- Sys.time()
    time <- end_time - start_time
    output <- list("time" = time, "data" = data)
  } # end GP2
  
  if (!is.null(init)){
    output$data <- output$data[(order+1):length(output$data)]
  }

  return(output)
}

#' @title installJuliaPackages
#' @description checks for needed \proglang{julia} packages and installs them if not installed.
#' @importFrom JuliaConnectoR juliaEval
#' @return no return value, called to install \proglang{julia} packages in \proglang{julia}.
#' @export
installJuliaPackages <- function(){
  
  strings1 <- c('"Coconots"')
  strings2 <- c('Pkg.add("Coconots")')
  for (i in 1:length(strings1)){
    if (!JuliaConnectoR::juliaEval(paste0(strings1[i], ' in keys(Pkg.project().dependencies)'))){
      #JuliaConnectoR.utils::install_julia_packages(strings2[i])
      JuliaConnectoR::juliaEval("using Pkg")
      JuliaConnectoR::juliaEval(paste0(strings2[i]) )
    }
  }
  JuliaConnectoR::juliaEval('
      using Coconots
      using Random
    ')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cocoSoc.R
\name{cocoSoc}
\alias{cocoSoc}
\title{Computes Scores for Various Models Maintaining a Common Sample}
\usage{
cocoSoc(
  data,
  models = "all",
  print.progress = TRUE,
  max_x_score = 50,
  julia = FALSE,
  ...
)
}
\arguments{
\item{data}{A numeric vector containing the data to be used for modeling}

\item{models}{A character string specifying which models to use. Default is \code{"all"}, which uses both Poisson and GP models.}

\item{print.progress}{A logical value indicating whether to print progress messages (Default: \code{TRUE}).}

\item{max_x_score}{An integer which is used as the maximum count for the computation
of the score (defaul: \code{50})}

\item{julia}{if TRUE, \code{cocoSoc} is run with \proglang{julia} (default: FALSE)}

\item{...}{Additional arguments to be passed to the \code{cocoReg} function.}
}
\value{
A list of class \code{"cocoSoc"} containing:
\describe{
\item{fits}{A list of fitted model objects.}
\item{scores_list}{A list of score objects for each model.}
\item{scores_df}{A data frame containing the logarithmic, quadratic, and ranked probability scores for each model.}
}
}
\description{
This function computes log, quadrtic and ranked probability scores for Poisson and Generalized Poisson models.
}
\details{
Supports model selection by computing score over a range of models while maintaining a common sample and
a common specification.
}
\author{
Manuel Huth
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.coco.R
\name{predict.coco}
\alias{predict.coco}
\title{K-Step Ahead Forecast Distributions}
\usage{
\method{predict}{coco}(
  object,
  k = 1,
  number_simulations = 1000,
  alpha = 0.05,
  simulate_one_step_ahead = FALSE,
  max = NULL,
  epsilon = 1e-08,
  xcast = NULL,
  decimals = 4,
  julia = FALSE,
  ...
)
}
\arguments{
\item{object}{An object that has been fitted previously, of class coco.}

\item{k}{The number of steps ahead for which the forecast should be computed (Default: 1).}

\item{number_simulations}{The number of simulation runs to compute (Default: 1000).}

\item{alpha}{Significance level used to construct the prediction intervals (Default: 0.05).}

\item{simulate_one_step_ahead}{If FALSE, the one-step ahead prediction is obtained using the analytical predictive distribution. If TRUE, bootstrapping is used.}

\item{max}{The maximum number of the forecast support for the plot. If NULL all values for which the cumulative distribution function is below 1- epsilon are used for the plot.}

\item{epsilon}{If max is NULL, epsilon determines the range of the support that is used by subsequent automatic plotting using R's plot() function.}

\item{xcast}{An optional matrix of covariate values for the forecasting. If \code{NULL}, the function assumes no covariates.}

\item{decimals}{Number of decimal places for the forecast probabilities}

\item{julia}{if TRUE, the estimate is predicted with \proglang{julia} (Default: FALSE).}

\item{...}{Optional arguments.}
}
\value{
A \code{cocoForecastCollection} object (list of \code{cocoForecast} objects, one per step ahead). Use \code{summary()} to obtain a data frame of point forecasts (mean, median, mode) and prediction intervals. Individual steps are accessible via \code{forecast[[i]]}.
}
\description{
Computes the k-step ahead forecast (distributions) using the models in the coconots package.
}
\details{
Returns forecasts for each mass point of the k-step ahead
distribution for the fitted model. The exact predictive distributions for
one-step ahead predictions for
the models included here are provided in Jung and Tremayne (2011), maximum
likelihood estimates replace the true model parameters.
For k>1 forecast distributions are estimated using a parametric bootstrap. See Jung and Tremanye (2006).
Out-of-sample values
for covariates can be provided, if necessary.

for k > 1
}
\examples{
length <- 500
pars <- c(1, 0.4)
set.seed(12345)
data <- cocoSim(order = 1, type = "Poisson", par = pars, length = length)
fit <- cocoReg(order = 1, type = "Poisson", data = data)
forecast <- predict(fit, k=1, simulate_one_step_ahead = FALSE)
plot(forecast[[1]]) #plot one-step ahead forecast distribution

}
\references{
Jung, R.C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 3, 268--280.

Jung, R.C. and Tremayne, A.R. (2006) Coherent forecasting in integer time series models.
\emph{International Journal of Forecasting} \bold{22}, 223--238
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cocoBoot.R
\name{cocoBoot}
\alias{cocoBoot}
\title{Bootstrap Based Model Assessment Procedure}
\usage{
cocoBoot(
  coco,
  numb.lags = 21,
  rep.Bootstrap = 1000,
  conf.alpha = 0.05,
  julia = FALSE,
  julia_seed = NULL
)
}
\arguments{
\item{coco}{An object of class coco}

\item{numb.lags}{Number of lags for which to compute sample autocorrelations (default: 21).}

\item{rep.Bootstrap}{Number of bootstrap replicates to use (default: 1000)}

\item{conf.alpha}{\eqn{100(1-\code{conf.alpha})\%} probability interval for the acceptance envelopes (default: 0.05)}

\item{julia}{if TRUE, the bootstrap is run with \proglang{julia} (default: FALSE)}

\item{julia_seed}{Seed for the \proglang{julia} implementation. Only used if \proglang{julia} equals TRUE}
}
\value{
an object of class cocoBoot. It contains the bootstrapped confidence intervals
of the autocorrelations and information on the model specifications.
}
\description{
Model checking procedure emphasising reproducibility in fitted models, as proposed by Tsay (1992).
}
\details{
Bootstrap-generated acceptance envelopes for the autocorrelation function provides an overall evaluation by comparing it with the sample autocorrelation function in a joint plot.
}
\examples{
lambda <- 1
alpha <- 0.4
set.seed(12345)
data <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)
fit <- cocoReg(order = 1, type = "Poisson", data = data)

# bootstrap model assessment - R implementation
boot_r <- cocoBoot(fit, rep.Bootstrap=400)
plot(boot_r)
}
\references{
Tsay, R. S. (1992) Model checking via parametric bootstraps in time series analysis. \emph{Applied Statistics} \bold{41}, 1--15.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cocoPit.R
\name{cocoPit}
\alias{cocoPit}
\title{Probability Integral Transform Based Model Assessment Procedure}
\usage{
cocoPit(coco, J = 10, conf.alpha = 0.05, julia = FALSE)
}
\arguments{
\item{coco}{An object of class coco}

\item{J}{Number of bins for the histogram (default: 10)}

\item{conf.alpha}{Significance level for the confidence intervals (default: 0.05)}

\item{julia}{if TRUE, the PIT is computed with \proglang{julia} (default: FALSE)}
}
\value{
an object of class cocoPit. It contains the probability integral
transform values,  p-value of the chi-square goodness of fit test and information on the model specifications.
}
\description{
Computes the probability integral transform (PIT) and provides
the non-randomized PIT histogram for assessing absolute performance of a
fitted model as proposed by Czado et al. (2009).
}
\details{
The adequacy of a distributional assumption for a model is assessed by
checking the cumulative non-randomized PIT distribution for uniformity.
A useful graphical device is the PIT histogram, which displays this
distribution to J equally spaced bins. We supplement the graph by
incorporating approximately \eqn{100(1 - \alpha)\%} confidence intervals obtained
from a standard chi-square goodness-of-fit test of the null hypothesis that
the J bins of the histogram are drawn from a uniform distribution.
For details, see Jung, McCabe and Tremayne (2016).
}
\examples{
lambda <- 1
alpha <- 0.4
set.seed(12345)
data <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)
fit <- cocoReg(order = 1, type = "Poisson", data = data)

#PIT R implementation
pit_r <- cocoPit(fit)
plot(pit_r)
}
\references{
Czado, C., Gneiting, T. and Held, L. (2009) Predictive model assessment for count data. \emph{Biometrics} \bold{65}, 1254--61.

Jung, R. C., McCabe, B.P.M. and Tremayne, A.R. (2016). Model validation and diagnostics. \emph{In Handbook of Discrete
Valued Time Series}. Edited by Davis, R.A., Holan, S.H., Lund, R. and Ravishanker, N.. Boca Raton: Chapman and
Hall, pp. 189--218.

Jung, R. C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 3, 268--280.
}
\author{
Manuel Huth
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/downloads.R
\docType{data}
\name{downloads}
\alias{downloads}
\title{Time Series of Daily Downloads of a TeX-Editor}
\format{
A time series (\code{ts}) object containing daily data from June 2006 to February 2007.
}
\source{
Weiss, C.H. (2008)
Thinning operations for modelling time series of counts -- a survey.
\emph{Advances in Statistical Analysis}, \bold{92}, 319--341.
}
\usage{
downloads
}
\description{
The data represent the number of daily downloads of a TeX-editor between
June 2006 and February 2007. The dataset contains 267 observations.
The data have been provided by Christian Weiss.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/installJuliaPackages.R
\name{installJuliaPackages}
\alias{installJuliaPackages}
\title{installJuliaPackages}
\usage{
installJuliaPackages()
}
\value{
no return value, called to install \proglang{julia} packages in \proglang{julia}.
}
\description{
checks for needed \proglang{julia} packages and installs them if not installed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cocoScore.R
\name{cocoScore}
\alias{cocoScore}
\title{Scoring Rule Based Model Assessment Procedure}
\usage{
cocoScore(coco, max_x = 50, julia = FALSE)
}
\arguments{
\item{coco}{An object of class coco}

\item{max_x}{An integer which is used as the maximum count for the computation
of the score (default: \code{50})}

\item{julia}{if TRUE, the scores are computed with \proglang{julia} (default: FALSE).}
}
\value{
a list containing the log score, quadratic score and ranked probability score.
}
\description{
The function computes log, quadratic and ranked probability scores for assessing relative performance of a fitted model.
}
\details{
Scoring rules assign a numerical score based on the predictive
distribution and the observed data  to measure the quality of probabilistic predictions.
They are provided here as a model selection tool and are computed as
averages over the relevant set of (in-sample) predictions. Scoring rules are, generally, negatively oriented
penalties that one seeks to minimize. The literature has developed a large number of scoring
rules and, unless there is a unique and clearly defined underlying decision problem,
there is no automatic choice of a (proper) scoring rule to be used in any given situation.
Therefore, the use of a variety of scoring rules may be appropriate to take advantage of
specific emphases and strengths. Three proper scoring rules
(for a definition of the concept of propriety see Gneiting and Raftery, 2007),
which Jung, McCabe and Tremayne (2016) found to be particularly useful, are implemented.
For more information see the references listed below.
}
\examples{
lambda <- 1
alpha <- 0.4
set.seed(12345)
data <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)
fit <- cocoReg(order = 1, type = "Poisson", data = data)

# scoring rules - R implementation
score_r <- cocoScore(fit)
}
\references{
Czado, C. and Gneitling, T. and Held, L. (2009) Predictive Model Assessment for Count Data. \emph{Biometrics}, \bold{65},  1254--1261.

Gneiting, T. and Raftery, A. E. (2007) Strictly proper scoring rules, prediction, and estimation. \emph{Journal
of the American Statistical Association}, 102:359-378.

Jung, R. C., McCabe, B.P.M. and Tremayne, A.R. (2016). Model validation and diagnostics. \emph{In Handbook of Discrete
Valued Time Series}. Edited by Davis, R.A., Holan, S.H., Lund, R. and Ravishanker, N.. Boca Raton: Chapman and
Hall, pp. 189--218.

Jung, R. C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 268--280.
}
\author{
Manuel Huth
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cuts.R
\docType{data}
\name{cuts}
\alias{cuts}
\title{Time Series of Monthly Counts of Claimants Collecting Wage Loss Benefit}
\format{
A time series (\code{ts}) object containing monthly data from January 1985 to December 1994.
}
\source{
Freeland, R. K.; McCabe, B.P.M. (2004)
Analysis of count data by means of the Poisson autoregressive model.
\emph{Journal of Time Series Analysis}, \bold{25}, 701--722.
}
\usage{
cuts
}
\description{
Monthly counts of claimants collecting wage loss benefit for injuries in the workplace
at one specific service delivery location of the Workers Compensation Board
of British Columbia, Canada in the period January 1985 to December 1994.
Only injuries due to cuts and lacerations are considered.
The data have been provided by Brendan McCabe.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/goldparticle.R
\docType{data}
\name{goldparticle}
\alias{goldparticle}
\title{Time Series of Gold Particle Counts in a Well-Defined Colloidal Solution}
\format{
A time series (\code{ts}) object containing 370 observations at equidistant time points.
}
\source{
Jung, R.C.; Tremayne, A.R. (2006)
Coherent forecasting in integer time series models.
\emph{International Journal of Forecasting}, \bold{22}, 223--238.

Westgren, A. (1916) Die Veraenderungsgeschwindigkeit der lokalen
Teilchenkonzentration in kolloidalen Systemen (Erste Mitteilung).
\emph{Arkiv foer Matematik, Astronomi och Fysik}, \bold{11}, 1--24.
}
\usage{
goldparticle
}
\description{
A sample of 370 counts of gold particles in a well-defined colloidal
solution at equidistant points in time. The data were originally
published in Westgren (1916) and later used in Jung and Tremayne (2006).
}
\examples{
plot(goldparticle)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setJuliaSeed.R
\name{setJuliaSeed}
\alias{setJuliaSeed}
\title{Set Seed for \proglang{julia}'s Random Number Generator}
\usage{
setJuliaSeed(julia_seed)
}
\arguments{
\item{julia_seed}{An integer seed value to be passed to \proglang{julia}'s random number generator.}
}
\description{
Sets the seed for \proglang{julia}'s random number generator to ensure reproducibility.
}
\details{
This function initializes the necessary \proglang{julia} functions and sets the random seed for \proglang{julia}.
If the provided seed is NULL, the function does nothing.
}
\author{
Manuel Huth
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coconots-package.R
\docType{package}
\name{coconots-package}
\alias{coconots}
\alias{coconots-package}
\title{coconots: Convolution-Closed Models for Count Time Series}
\description{
Useful tools for fitting, validating, and forecasting of practical convolution-closed time series models for low counts are provided. Marginal distributions of the data can be modelled via Poisson and Generalized Poisson innovations. Regression effects can be incorporated through time varying innovation rates. The models are described in Jung and Tremayne (2011) \doi{10.1111/j.1467-9892.2010.00697.x} and the model assessment tools are presented in Czado et al. (2009) \doi{10.1111/j.1541-0420.2009.01191.x} and, Tsay (1992) \doi{10.2307/2347612}.
}
\author{
\strong{Maintainer}: Manuel Huth \email{manuel.huth@yahoo.com}

Authors:
\itemize{
  \item Robert C. Jung
  \item Andy Tremayne
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cocoResid.R
\name{cocoResid}
\alias{cocoResid}
\title{Residual Based Model Assessment Procedure}
\usage{
cocoResid(coco, val.num = 1e-10)
}
\arguments{
\item{coco}{An object of class "coco}

\item{val.num}{A non-negative real number that halts the calculation once the cumulative probability reaches 1-\code{val.num}}
}
\value{
a list that includes the (Pearson) residuals, conditional expectations, conditional variances,
and information on the model specifications.
}
\description{
Calculates the (Pearson) residuals of a fitted model for model evaluation purposes.
}
\details{
The Pearson residuals are computed as the scaled
deviation of the observed count from its conditional expectation given the relevant
past history, including covariates, if applicable. If a fitted model is correctly specified,
the Pearson residuals should exhibit mean zero, variance one, and no significant serial correlation.
}
\author{
Manuel Huth
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cocoSim.R
\name{cocoSim}
\alias{cocoSim}
\title{Simulation of Count Time Series}
\usage{
cocoSim(
  type,
  order,
  par,
  length,
  xreg = NULL,
  init = NULL,
  julia = FALSE,
  julia_seed = NULL,
  link_function = "log"
)
}
\arguments{
\item{type}{character, either "Poisson" or "GP" indicating the type of the innovation distribution}

\item{order}{integer, either 1 or 2 indicating the order of the model}

\item{par}{numeric vector, the parameters of the model, the number of elements
in the vector depends on the type and order specified.}

\item{length}{integer, the number of observations in the generated time series}

\item{xreg}{data frame of control variables (defaul: NULL)}

\item{init}{numeric vector, initial data to use (default: NULL). See details
for more information on the usage.}

\item{julia}{If TRUE, the \proglang{julia} implementation is used. In this case, init is ignored but it might be faster (default: FALSE).}

\item{julia_seed}{Seed for the \proglang{julia} implementation. Only used if \proglang{julia} equals TRUE.}

\item{link_function}{Specifies the link function for the conditional mean of the innovation (\eqn{\lambda}). The default is \code{log}, but other available options include \code{identity} and \code{relu}. This parameter is applicable only when covariates are used. Note that using the \code{identity} link function may result in \eqn{\lambda} becoming negative. To prevent this, ensure all covariates are positive and restrict the parameter \eqn{\beta} to positive values.}
}
\value{
a vector of the simulated time series
}
\description{
The function generates a time series of low counts from the (G)PAR model class for a specified
innovation distribution, sample size, lag order,
and parameter values.
}
\details{
The function checks for valid input of the type, order, parameters, and initial data
before generating the time series.

The init parameter allows users to set a custom burn-in period
for the simulation. By default, when simulating with covariates, no burn-in
period is specified since there is no clear choice on the covariates.
However, the init argument gives users the flexibility to select an
appropriate burn-in period for the covariate case. One way to do this is to
simulate a time series using \code{\link{cocoSim}} with appropriate covariates and pass the
resulting time series to the
init argument of a new \code{\link{cocoSim}} run so that the first time series is used as
the burn-in period.
If init is not specified for the covariate case, a warning will be returned
to prompt the user to specify a custom burn-in period. This helps ensure that
the simulation accurately captures the dynamics of the system being modeled.
}
\examples{
#First Order Model
lambda <- 1
alpha <- 0.4
set.seed(12345)

# Simulate using the RCPP implementation
data_rcpp <- cocoSim(order = 1, type = "Poisson", par = c(lambda, alpha), length = 100)

# Second Order Model
lambda <- 1
alpha_1 <- 0.3
alpha_2 <- 0.1
alpha_3 <- 0.2
eta <- 0.2
data_j <- cocoSim(order = 2, type = "GP",
                  par = c(lambda, alpha_1, alpha_2,alpha_3,eta),
                  length = 100)
}
\author{
Manuel Huth
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cocoReg.R
\name{cocoReg}
\alias{cocoReg}
\title{Fitting First- and Second Order (G)PAR Models}
\usage{
cocoReg(
  type,
  order,
  data,
  xreg = NULL,
  constrained.optim = TRUE,
  b.beta = -10,
  start = NULL,
  start.val.adjust = TRUE,
  method_optim = "Nelder-Mead",
  replace.start.val = 1e-05,
  iteration.start.val = 0.6,
  method.hessian = "Richardson",
  cores = 2,
  julia = FALSE,
  julia_installed = FALSE,
  link_function = "log"
)
}
\arguments{
\item{type}{character, either "Poisson" or "GP" indicating the type of the innovation distribution}

\item{order}{integer, either 1 or 2 indicating the order of the model}

\item{data}{time series data to be used in the analysis}

\item{xreg}{optional matrix of explanatory variables (without constant term) for use in a regression model}

\item{constrained.optim}{logical indicating whether optimization should be constrained, currently only available in the R version}

\item{b.beta}{numeric value indicating the lower bound for the parameters of the explanatory variables for the optimization, currently only available in the \proglang{R} version}

\item{start}{optional numeric vector of starting values for the optimization}

\item{start.val.adjust}{logical indicating whether starting values should be adjusted, currently only available in the R version}

\item{method_optim}{character string indicating the optimization method to be used, currently only available in the R version. In the \proglang{julia} implementation this is by default the LBFGS algorithm}

\item{replace.start.val}{numeric value indicating the value to replace any invalid starting values, currently only available in the \proglang{R} version}

\item{iteration.start.val}{numeric value indicating the proportion of the interval to use as the new starting value, currently only available in the \proglang{R} version}

\item{method.hessian}{character string indicating the method to be used to approximate the Hessian matrix, currently only available in the \proglang{R} version}

\item{cores}{numeric indicating the number of cores to use, currently only available in the \proglang{R} version (default: 2)}

\item{julia}{if TRUE, the model is estimated with \proglang{julia}. This can improve computational speed significantly since \proglang{julia} makes use of derivatives using autodiff. In this case, only \code{type}, \code{order}, \code{data}, \code{xreg}, and \code{start} are used as other inputs (default: FALSE).}

\item{julia_installed}{if TRUE, the model \proglang{R} output will contain a \proglang{julia} compatible output element.}

\item{link_function}{Specifies the link function for the conditional mean of the innovation (\eqn{\lambda}). The default is \code{log}, but other available options include \code{identity}, \code{relu}, and \code{softplus}. This parameter is applicable only when covariates are used. Note that using the \code{identity} link function may result in \eqn{\lambda} becoming negative. To prevent this, ensure all covariates are positive and restrict the parameter \eqn{\beta} to positive values by setting \code{b.beta} to a small positive value.}
}
\value{
an object of class coco. It contains the parameter estimates, standard errors, the log-likelihood,
and information on the model specifications. If \proglang{julia} is used for parameter estimation or the \proglang{julia} installation
parameter is set to TRUE, the results contain an additional Julia element that is called from the model \proglang{julia}
assessment tools if they are run with the \proglang{julia} implementation.
}
\description{
The function fits first- and second-order (Generalized) Poisson
integer autoregressive \code{(G)PAR} time series models for count data as discussed in Jung and Tremayne (2011). Autoregressive dependence on past counts is modeled using a special random operator that preserves integer values and, through closure under convolution, ensures that the marginal distribution remains within the same family as the innovations.

These models can be viewed as stationary finite-order Markov chains, where the innovation distribution can be either Poisson or Generalized Poisson, the latter accounting for overdispersion. Estimation is performed via maximum likelihood, with an option to impose linear constraints. Without constraints, parameters may fall outside the theoretically feasible space, but optimization may be faster.

Method of moments estimators are used to initialize numerical optimization, though custom starting values can be provided. If \proglang{julia} is installed, users can opt to run the optimization in \proglang{julia} for potentially faster computation and improved numerical stability via automatic differentiation. See below for details on the \proglang{julia} implementation.
}
\details{
Let a time series of counts be \eqn{\{X_t\}} and be \eqn{R(\cdot)} a random operator that differs between model specifications.
For more details on the random operator, see Jung and Tremayne (2011) and Joe (1996).
The general first-order model is of the form
\deqn{X_t = R(X_{t-1}) + W_t,}
and the general second-order model of the form
\deqn{X_t = R(X_{t-1}, X_{t-2}) + W_t,}
where \eqn{W_t} are i.i.d Poisson (\eqn{W_t \sim Po(\lambda_t)}) or Generalized
Poisson (\eqn{W_t \sim GP(\lambda_t, \eta)}) innovations. Through closure under convolution
the marginal distributions of \eqn{\{X_t\}} are therefore Poisson or Generalized Poisson distributions, respectively.

If no covariates are used \eqn{\lambda_t = \lambda} and if covariates are used
\deqn{g(\lambda_t) = \left(\beta_0 + \sum_{j = 1}^k \beta_j \cdot z_{t,j} \right),}
whereby \eqn{z_{t,j}} is the \eqn{j}-th covariate at time \eqn{t} and \eqn{g} is a link function.
Currently supported link functions are: the identity \eqn{g(x) = x}, the logarithmic link
\eqn{g(x) = \ln x}, relu \eqn{g(x) = \max(x, \epsilon)} for small \eqn{\epsilon > 0}, and softplus
\eqn{g(x) = \log(\exp(x) - 1)}. To ensure positivity of \eqn{\lambda} if the identity function is used, \eqn{\beta_j, z_{t,j} > 0} must be enforced.
Alternatively, computational values of \eqn{\lambda \leq 0} can be set to a small positive value.
This option is named 'relu', due to its similarity to a ReLU function commonly used in machine learning.
The softplus link uses a numerically stable implementation of its inverse \eqn{\log(1 + \exp(x))}, ensuring \eqn{\lambda > 0} and smooth gradients everywhere.

Standard errors are computed by the square root of the diagonal elements of the inverse Hessian.

This function is implemented in two versions. The default runs on RCPP.
An alternative version uses a \proglang{julia} implementation which can be chosen by
setting the argument \proglang{julia} to TRUE. In order to
use this feature, a running \proglang{julia} installation is required on the system.
The RCPP implementation uses the derivative-free Nelder-Mead optimizer
to obtain parameter estimates. The \proglang{julia} implementation makes use of \proglang{julia}'s
automatic differentiation in order to obtain gradients such that it can use the LBFGS algorithm for
optimization. This enhances the numeric stability of the optimization and yields
an internal validation if both methods yield qualitatively same parameter estimates.
Furthermore, the \proglang{julia} implementation can increase the computational speed
significantly, especially for large models.

The model assessment tools \code{\link{cocoBoot}}, \code{\link{cocoPit}}, and \code{\link{cocoScore}}
will use a \proglang{julia} implementation as well, if the \code{\link{cocoReg}} was run with \proglang{julia}.
Additionally, one can make the RCPP output of \code{\link{cocoReg}} compatible with the \proglang{julia}
model assessments by setting \code{julia_installed} to true. In this case, the user can choose
between the \pkg{RCPP} and the \proglang{julia} implementation for model assessment.
}
\examples{
## GP2 model without covariates
length <- 1000
par <- c(0.5,0.2,0.05,0.3,0.3)
data <- cocoSim(order = 2, type = "GP", par = par, length = length)
fit <- cocoReg(order = 2, type = "GP", data = data)

##Poisson1 model with covariates
length <- 1000
period <- 12
sin <- sin(2*pi/period*(1:length))
cos <- cos(2*pi/period*(1:length))
cov <- cbind(sin, cos)
par <- c(0.2, 0.2, -0.2)
data <- cocoSim(order = 1, type = "Poisson", par = par, xreg = cov, length = length)
fit <- cocoReg(order = 1, type = "Poisson", data = data, xreg = cov)
}
\references{
Jung, R. C. and Tremayne, A. R. (2011) Convolution-closed models for count time series with applications. \emph{Journal of Time Series Analysis}, \bold{32}, 268--280.

Joe, H. (1996) Time series models with univariate margins in the convolution-closed infinitely divisible class. \emph{Journal of Applied Probability}, \bold{33}, 664--677.
}
\author{
Manuel Huth
}
